"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/ui/lib/__init__.py $
$Id: __init__.py 29247 2007-01-29 22:12:28Z dbinger $
"""
from qpy.compile import compile_qpy_files
compile_qpy_files(__path__[0])



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/comment.py $
$Id: comment.py 29510 2007-02-27 22:15:17Z dbinger $
"""
from dulcinea.base import DulcineaPersistent
from dulcinea.timestamped import Timestamped
from dulcinea.user import DulcineaUser
from qp.lib.spec import specify, add_getters, spec, string

class Comment(DulcineaPersistent, Timestamped):

    text_is = spec(
        string)
    user_is = spec(
        (None, DulcineaUser),
        "The author of the comment")

    def __init__ (self, text, user):
        Timestamped.__init__(self)
        specify(self, text=text, user=user)

    def __str__ (self):
        return "%s at %s" % (self.user, self.get_timestamp())


add_getters(Comment)



#!/www/python/bin/python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/site/update_db.py $
$Id: update_db.py 29993 2007-08-13 17:00:28Z dbinger $

When we update the live site, this is our procedure.
Make a tarball of the running code and database file.
This is the emergency kit.
Copy the revised code into place.

If the code revisions require corresponding data revisions, 
  Copy the database file to a temporary name.
  Run an update_db.py script that changes attribute names, classes,
  or whatever is needed to convert the database to agree with the 
  new version of the code.  
  qp -d <site>
  Move the old database file away.  Move the fixed version into place.
  qp -u <site>
Otherwise,
  there are no data changes, so we just 
  qp -q <site>

This script is intended to serve as an example of an update_db.py
for a site named "dulcinea".
"""
from qp.lib.site import Site
import sys
connection = root = commit = pack = trace = None
# This call sets connection, root, commit, pack, and trace.
Site('dulcinea').update_db_environment(globals(), *sys.argv[1:])

# Do required changes here.


commit() 
# pack() # optional, but nice if you have time for it.  



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_outline.py $
$Id: utest_outline.py 27434 2005-09-16 15:02:48Z dbinger $
"""
from dulcinea.outline import Outline
from sancho.utest import UTest, raises

def outline_from_nested_lists(nested_lists):
    if not isinstance(nested_lists, list):
        return nested_lists
    return Outline([outline_from_nested_lists(x)
                    for x in nested_lists])

class OutlineTest(UTest):

    def a(self):
        x = Outline()
        raises(ValueError, x.index, 'b')
        x.extend(['a', 'b', 'b', 'c'])
        assert x.index('b') == 1, x.index('b')
        x[:] = ['a', Outline(['c', 'b']), 'd']
        assert x.index('b') == [1, 1]
        assert 'b' == x[[1, 1]]
        x[[1,1]] = 'e'
        assert 'e' == x[[1, 1]]
        raises(ValueError, x.index, 'b')
        raises(IndexError, x.__getitem__, 100)
        x[0] = 'j'
        assert x[2] == 'd', x.data
        del x[-1]
        raises(ValueError, x.index, 'd')

    def b(self):
        x = outline_from_nested_lists(['a', 'b', ['c', [['d']]], ['e'], []])
        assert x.get('3.1') == 'c'
        assert x.get([2, 0]) == 'c'
        for letter in 'abcde':
            assert x.get(x.index(letter)) == letter
            assert x.get(x.index_to_string(x.index(letter))) == letter
        keys = list(x.iterkeys())
        values = list(x.itervalues())
        assert keys == [k for k, v in x.iteritems()]
        assert values == [v for k, v in x.iteritems()]
        x.insert([2, 0], 'f')
        assert x.get('3.1') == 'f'
        assert x.get('3.2') == 'c'
        assert x.get('3.2.3.4') == None
        v =  x[[1]]
        del x[[1]]
        raises(ValueError, x.index, v)
        raises(IndexError, x.__setitem__, [0, 1], 'a')

    def c(self):
        x = Outline()
        assert x.get(0, 1) == 1

if __name__ == '__main__':
    OutlineTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_links.py $
$Id: utest_links.py 31152 2008-09-18 11:27:09Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.contact import Contact
from dulcinea.links import LinkItem, LinkFolder, get_lower_title


class AdminContact(Contact):

    def is_admin(self):
        return True


class LinksTest (CommonTest):

    def check_link(self):
        self.link = LinkItem()
        self.link.set_link_url('http://www.example.org')
        self.link.get_link_url()
        self.link.set_text('hi')
        assert self.link.get_text() == 'hi'
        self.link.set_email('hi')
        assert self.link.get_email() == 'hi'
        self.link.as_text()
        self.link.as_email()
        assert self.link.is_approved() == False
        self.link.set_approved(True)
        assert self.link.is_approved() == True
        self.link.get_allowed_mime_types()
        self.link.get_allowed_mime_types(user=AdminContact('superuser'))
        assert self.link.get_local_url() == '/links/None/'
        self.link2 = LinkItem()
        [self.link, self.link2].sort(key=get_lower_title)

    def check_link_folder(self):
        self.link_db = LinkFolder()
        self.link_db.get_category('foo')
        self.link_db.links_with_category('foo')


if __name__ == "__main__":
    LinksTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_material.py $
$Id: utest_material.py 27147 2005-08-03 13:59:08Z dbinger $
"""
from dulcinea.material import Material, material_sort, MaterialDatabase
from sancho.utest import UTest


class MaterialTest (UTest):

    def check_materials_constructor (self):
        mdb = MaterialDatabase()
        assert list(mdb.get_materials()) == []
        m1 = Material('silicon')
        mdb.add_material(m1)
        assert list(mdb.get_materials()) == [m1]
        m2 = Material('silicon', 'Si')
        try:
            mdb.add_material(m2)
            assert 0
        except ValueError: pass
        assert list(mdb.get_materials()) == [m1]
        m2 = Material('silicon2', 'Si')
        mdb.add_material(m2)
        assert len(mdb.get_materials()) == 2
        m3 = Material('Oxygen', 'O<sub>2</sub>')
        mdb.add_material(m3)
        assert len(mdb.get_materials()) == 3
        assert len(mdb.get_material('Oxygen').html_color) == 7
        assert mdb.get_material('Oxygen').html_name == 'O<sub>2</sub>'
        assert mdb.get_material('Oxygen').name == 'Oxygen'
        m3 = Material('Oxygen2', 'O<sub>2</sub>', 'blue')
        mdb.add_material(m3)
        assert len(mdb.get_materials()) == 4

    def check_materials_database (self):
        mdb = MaterialDatabase()
        mdb.add_material(Material('silicon'))
        mdb.add_material(Material('silicon2', 'Si'))
        mdb.add_material(Material('Oxygen', 'O<sub>2</sub>'))
        mdb.add_material(Material('Oxygen2', 'O<sub>2</sub>','blue'))
        mdb.add_material(Material('gas'))
        assert len(mdb.get_material('gas').get_leaves()) == 1
        assert len(mdb.get_material('Oxygen').get_leaves()) == 1
        mdb.get_material('gas').add_child(mdb.get_material('Oxygen'))
        assert len(mdb.get_material('gas').get_leaves()) == 1
        assert len(mdb.get_material('gas').get_descendants()) == 1
        mdb.get_material('gas').add_child(mdb.get_material('Oxygen2'))
        assert len(mdb.get_material('gas').get_leaves()) == 2
        assert len(mdb.get_material('gas').get_descendants()) == 2
        mdb.get_material('gas').remove_child(mdb.get_material('Oxygen2'))
        assert len(mdb.get_material('gas').get_leaves()) == 1
        assert len(mdb.get_material('gas').get_descendants()) == 1

    def check_sort (self):
        material_list = []
        for (name, label) in (("chocolate_sauce", "chocolate sauce"),
                              ("peanut_butter", "peanut butter"),
                              ("1_2_3_foobaroxynol", "1,2,3-foobaroxynol"),
                              ("H2O", None)):
            m = Material(name)
            if label:
                m.set_label(label)
            material_list.append(m)

        sorted_list = [material_list[2],
                       material_list[0],
                       material_list[3],
                       material_list[1]]

        assert material_sort(list(material_list)) == sorted_list

        material_list.reverse()
        material_tuple = tuple(material_list)
        assert material_sort(material_tuple) == sorted_list

        def gen_materials ():
            for m in material_list:
                yield m

        materials = gen_materials()
        assert material_sort(materials) == sorted_list

if __name__ == "__main__":
    MaterialTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/property/property_template.py $
$Id: property_template.py 30630 2008-03-23 17:01:50Z dbinger $
"""
from dulcinea.base import DulcineaPersistent
from dulcinea.material import get_material_db
from dulcinea.physical_unit import get_standard_unit, PhysicalUnit
from dulcinea.physical_value import PhysicalValue
from qp.lib.spec import add_getters, string, pattern, both
from qp.lib.spec import boolean, instance, anything, spec, specify
from qpy import stringify
import dulcinea.property.property_type


# Default value for 'check_constraint()' method -- can't use None because
# it's a legitimate constraint.
_no_value = "_no_value"


class PropertyTemplate (DulcineaPersistent):
    """
    Property templates are used to guide property input.
    """
    constraint_is = spec(
        (None, list),
        "a list of allowed values for the property")

    description_is = spec(
        (None, string),
        "an accurate, concise, technical, readable description of the "
        "property")

    def __init__(self):
        assert self.__class__ is not PropertyTemplate, "abstract class"

    def get_description(self):
        return self.description

    def set_description(self, description):
        specify(self, description=description)

    def check_type(self, value):
        """(value : any)

        Ensure that 'value' is compatible with the property type.
        Raise PropertyTypeError if not.
        """
        self.get_type().check_value(value)

    def is_atomic(self):
        return self.get_type().is_atomic()

    def is_list(self):
        return self.get_type().is_list()

    def is_table(self):
        return self.get_type().is_table()

    def is_aggregate(self):
        return self.get_type().is_aggregate()

    def is_input(self):
        return 0

    def is_property(self):
        return 0

    def set_constraint(self, set):
        """(set : [property_value])

        Specify a constraint.  All elements of the constraint must be of the
        same type or a range, which must be the correct type for the
        property to which this constraint belongs.  For a value to meet the
        constraint, it must be equal to or within the range of one of the
        elements.
        """
        self.check_constraint(set)
        specify(self, constraint=set)

    def get_constraint(self):
        """Return the constraint or None if there is no constraint.
        """
        raise NotImplementedError

    def get_constraint_unit(self):
        constraint = self.get_constraint()
        if (constraint and isinstance(constraint[0], PhysicalValue)):
            return constraint[0].get_unit()
        else:
            return None

    def get_constraint_range(self):
        """() -> PhysicalValue

        Return the range of valid values described by the constraint.  If
        it's unconstrained, return None otherwise return the span of the
        set.
        """
        return self.get_type().get_constraint_range(self.get_constraint())

    def is_discrete(self):
        """Return true if this property can only take on discrete
        values.  If the constraint contains range values or there is no
        constraint then false is returned."""
        return self.get_type().is_discrete(self.get_constraint())

    def format(self, html=0):
        """Return a string with a brief summary of the property
        constraint, suitable to explain the constraint to a user before
        they supply a value.  Return None if constraint is empty.
        """
        return self.get_type().format_constraint(self.get_constraint(),
                                                 html=html)

    def explain_constraint(self, html=0):
        """Return a string with a fairly detailed explanation of the
        property constraint, suitable for use in error messages after
        user has violated the constraint.  Return None if constraint is
        empty.
        """
        constraint = self.get_constraint()
        return self.get_type().explain_constraint(constraint, html=html)

    def is_allowed_unit(self, unit):
        """Return true if 'unit' is an allowed unit for properties
        derived from this property template, false otherwise.  (If the
        'allowed_units' attribute is not defined, no units are allowed).
        """
        allowed_units = self.get_allowed_units()
        if allowed_units is None:
            return unit is None
        elif unit is None:
            return 0
        else:
            unit = get_standard_unit(unit)
            return unit in allowed_units

    def check_value(self, value):
        """(value : property_value)

        Checks if 'value' is of the allowed type and meets the constraint
        specified earlier with 'set_constraint()'.  Returns silently if
        no constraint was set, or if the value meets the constraint.
        Raises ConstraintError if a constraint is defined, but 'value'
        doesn't meet it.  Raises PropertyTypeError (indirectly) if the value
        is of the wrong type.
        """
        if value is None:
            return

        # Raises PropertyTypeError if type is wrong.
        self.check_type(value)

        constraint = self.get_constraint()

        if constraint is not None:
            self.get_type().check_value_in_constraint(value, constraint)

    def get_default_unit(self):
        return self.get_type().get_unit()

    def is_time(self):
        """
        Return true if this property is a PhysicalValue and the units are
        time units.
        """
        unit = self.get_default_unit()
        return unit and unit.is_time()

    def simplify(self):
        """() -> RangeValue | None
        Return a simplified version of this template's constraint (a
        RangeValue containing ints or floats). Return None if there is
        no constraint or if the constraint cannot be simplified.
        """
        property_type = self.get_type()
        if property_type.is_int() or property_type.is_physical_value():
            constraint_range = self.get_constraint_range()
            if constraint_range is not None:
                return property_type.simplify_value(constraint_range)
        return None

    def create_value(self, value=None):
        """Create a new Property based on this template."""
        from dulcinea.property.property import Property
        return Property(self, value)

    def __str__(self):
        constraint = self.get_constraint()
        if constraint:
            constraint = list(map(stringify, constraint))
            if len(constraint) > 5:
                constraint = (", ".join(constraint[0:3]) + " ... " +
                              constraint[-1])
            else:
                constraint = ", ".join(constraint)
            return "%s : set %s" % (self.get_name(), constraint)
        else:
            return "%s (unconstrained)" % self.get_name()


class MasterTemplate (PropertyTemplate):

    name_is = spec(
        both(string, pattern('^[a-z][a-z0-9_]*$')),
        "a software-friendly name for the property.")
    type_is = spec(
        instance('PropertyType'),
        "the type of this property")
    hidden_is = spec(
        boolean,
        "a UI hint to show the property less often")
    title_is = spec(
        (None, string),
        "the property name in a human-friendly form (if None, get_title "
        "derives the title from the name)")

    def __init__(self, name, property_type,
                 hidden=0,
                 description=None,
                 title=None,
                 constraint=None):
        self.set_name(name)
        if title is None:
            words = self.name.split("_")
            words[0] = words[0].capitalize()
            title = " ".join(words)
        specify(self,
                type=property_type,
                hidden=0,
                description=description,
                title=title,
                constraint=constraint)

    def get_allowed_units(self):
        units = self.get_type().get_compatible_units()
        if units:
            return [unit for unit in units if not unit.artificial]
        else:
            return None

    def get_master(self):
        return self

    def is_hidden(self):
        return self.hidden

    def set_name(self, name):
        specify(self, name=name)

    def set_hidden(self, hidden):
        specify(self, hidden=hidden)

    def set_title(self, title):
        specify(self, title=title)

    def get_constraint(self):
        """Return the constraint or None if there is no constraint.
        """
        if self.constraint is None and self.get_type().is_material():
            return list(get_material_db().get_materials())
        else:
            return self.constraint

    def create_input_template(self, constraint=None, description=None,
                              allowed_units=None, required=0,
                              default_value=None):
        """Create a new InputTemplate based on this template."""
        constraint = constraint or self.constraint
        return InputTemplate(self, constraint=constraint,
                             description=description,
                             allowed_units=allowed_units,
                             required=required,
                             default_value=default_value)

    def check_constraint(self, set=_no_value):
        if set is _no_value:
            set = self.constraint
        self.get_type().check_constraint(set)

    def is_required(self):
        return 0


add_getters(MasterTemplate)


class InputTemplate (PropertyTemplate):
    """
    Input templates are created to guide input.
    """
    template_is = spec(
        MasterTemplate,
        "the master template for this template; this must always be set")
    allowed_units_is = spec(
        (None, [PhysicalUnit]),
        "a list of PhysicalUnit instances that controls what unit may be "
        "used when entering a property.")
    required_is = spec(
        boolean,
        "true if a value is required for this property")
    default_value_is = spec(
        anything,
        "A typical value for this property")

    def __init__(self, template, constraint=None, description=None,
                 allowed_units=None, required=0, default_value=None):
        specify(self,
                template=template,
                default_value=default_value,
                description=description,
                required=required)
        self.set_allowed_units(allowed_units)
        self.set_constraint(constraint)

    def check_constraint(self, set=_no_value):
        """Verify that set is a valid constraint for this template.
        """
        if set is None:
            return
        if set is _no_value:
            set = self.constraint
        self.get_type().check_set_in_constraint(
            set,
            self.template.get_constraint())

    def get_template(self):
        return self.template

    def get_master(self):
        return self.template.get_master()

    def get_name(self):
        return self.template.get_name()

    def set_default_value(self, value):
        self.check_value(value)
        self.default_value = value

    def check_default_value(self):
        self.check_value(self.default_value)

    def get_default_value(self):
        if self.default_value == '':
            return None
        return self.default_value

    def is_hidden(self):
        return self.get_master().is_hidden()

    def is_required(self):
        return self.required

    def is_input(self):
        return 1

    def get_title(self):
        return self.template.get_title()

    def get_description(self):
        return self.description or self.template.get_description()

    def get_type(self):
        return self.template.get_type()

    def set_allowed_units(self, units):
        """(units : [PhysicalUnit | string] | None)"""
        if not units or units == [None]:
            self.allowed_units = None
        else:
            assert self.get_type().is_physical_value(), (
                'allowed_units only apply to physical value properties '
                'not %r' % self.get_type())
            self.allowed_units = [get_standard_unit(unit)
                                  for unit in units]

    def get_allowed_units(self):
        return (self.allowed_units or self.template.get_allowed_units())

    def get_constraint(self):
        """Return the constraint or None if there is no constraint.
        """
        if self.constraint is None:
            return self.template.get_constraint()
        else:
            return self.constraint

    def set_required(self, required):
        specify(self, required=required)

    def create_default_value(self):
        """Create a new Property based on this template."""
        from dulcinea.property.property import Property
        return Property(self, self.get_default_value())




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/ui/page_loader.py $
$Id: page_loader.py 31181 2008-09-25 20:26:26Z rmasse $
"""
from dulcinea.common import set_test_environment
from durus.utils import BytesIO, as_bytes
from pprint import pformat
from qp.pub.common import get_request, get_session, get_response, get_publisher
from qp.pub.publish import RespondNow
from sancho.utest import UTest
from subprocess import Popen, PIPE

class PageLoader (object):

    def set_test_environment(self, **kwargs):
        return set_test_environment(**kwargs)

    def __init__(self, user=None, site_name=None):
        self.set_test_environment(site_name=site_name)
        self.site_name = site_name
        session = get_session()
        for key, value in get_publisher().get_sessions().items():
            if value is session:
                self.session_key = key
                break
        else:
            assert 0, "no session!"
        if user is None:
            self.user = self.get_admin_user()
        else:
            self.user = user
        self.cookies = get_request().get_environ('HTTP_COOKIE')
        session.set_authenticated(self.user)
        self.session = session

    def process(self,
                PATH_INFO='/',
                input_string = '',
                SCRIPT_NAME='',
                SERVER_NAME='test@example.org',
                HTTPS='on',
                **kwargs):
        env = dict(PATH_INFO=PATH_INFO,
                   SCRIPT_NAME=SCRIPT_NAME,
                   SERVER_NAME=SERVER_NAME,
                   HTTPS=HTTPS,
                   **kwargs)
        get_publisher().process(BytesIO(as_bytes(input_string)), env)
        request = get_request()
        return request

    def get_admin_user(self):
        for user in get_publisher().get_users().itervalues():
            if user.is_admin():
                return user
        return None

    def get_paths(self, user):
        def get_paths_inner(directory, base_path):
            get_request().environ['PATH_INFO'] = base_path + '/'
            get_request().environ['SERVER_NAME'] = "example.org"
            get_request().scheme = 'https'
            self.session.set_authenticated(user)
            get_request().session = self.session
            exports = []
            try:
                for export in directory.get_exports():
                    exports.append(export[0])
            except RespondNow:
                pass
            exports = sorted(exports)
            for name in exports:
                objname = directory._q_translate(name)
                if objname is None:
                    obj = directory._q_lookup(name)
                else:
                    obj = getattr(directory, objname)
                if hasattr(obj, '_q_traverse'):
                    for path in get_paths_inner(obj, base_path + '/' + name):
                        yield path
                else:
                    yield base_path +'/' + name
        set_test_environment(site_name=self.site_name)
        self.session.set_authenticated(user)
        get_request().session = self.session
        return [path for path in get_paths_inner(
            get_publisher().get_root_directory(), '')]

    def get_processed_request(self, path, follow_redirects=(301,302), **kwargs):
        """
        Returns a request after processing, after one redirect if necessary.
        """
        if 'HTTP_COOKIE' not in kwargs:
            kwargs['HTTP_COOKIE'] = self.cookies
        self.session.set_authenticated(self.user)
        get_publisher().get_sessions()[self.session_key] = self.session
        request = self.process(path, **kwargs)
        n = 4
        while n > 0 and get_response().get_status_code() in follow_redirects:
            n -= 1
            location = get_response().get_header('location')
            path = '/' + location.split('/', 3)[-1] # strip http://server
            if '?' in path:
                path, query = path.split('?')
                kwargs['QUERY_STRING'] = query
            self.session.set_authenticated(self.user)
            get_publisher().get_sessions()[self.session_key] = self.session
            request = self.process(path, **kwargs)
        return request

    def is_reportable(self, line):
        if 'Warning: ' in line:
            warning = line.split('Warning: ', 1)[1].strip()
            if warning in (
                '<table> lacks "summary" attribute',
                '<a> attribute "href" lacks value',
                '<form> attribute "action" lacks value',
                '<input> attribute "name" lacks value',
                'trimming empty <option>',
                'trimming empty <span>',
                '<textarea> proprietary attribute "wrap"',
                '<a> cannot copy name attribute to id', # starts with digit
                ):
                return False
            if warning.startswith("trimming empty <"):
                return False
            if (warning.endswith('" already defined') and
                warning.startswith('<a> anchor "')):
                return False
        return True

    def expect(self, path, status=200, **kwargs):
        request = self.get_processed_request(path, **kwargs)
        try:
            assert get_response().get_status_code() == status, (
                "%r (expected %r) %s" % (get_response().get_status_code(),
                                         status,
                                         path))
        except:
            #import sys
            #get_response().write(sys.stdout)
            print("\n%s %s" % (get_response().get_status_code(), path))
            raise
        return (request, get_response())

    def validate_html(self, request, response, assert_valid=True):
        if response.get_mime_type() == 'text/html':
            s = BytesIO()
            for chunk in response.generate_body_chunks():
                s.write(chunk)
            body = s.get_value().decode('utf-8')
            subprocess = Popen("/usr/bin/env tidy -q -e 2>&1", stdin=PIPE, stdout=PIPE)
            ins = subprocess.stdout
            outs = subprocess.stdin
            outs.write(body)
            outs.close()
            body_lines = body.splitlines()
            errors = []
            for line in ins.readlines():
                if self.is_reportable(line):
                    splitline = line.split()
                    lineno = int(splitline[1])
                    colno = int(splitline[3])
                    if body_lines:
                        bodyline = body_lines[lineno-1]
                    else:
                        bodyline = ''
                    errors.append(
                        '%s    ->%s: %r\n\n' % (
                        line,
                        lineno,
                        bodyline[max(colno-20, 0):colno+100]))
            if errors:
                print("\nLoading %r:\n%s\n\n" % (
                    request.get_path(), ''.join(errors)))
                if assert_valid:
                    assert not errors, 'html errors found'
        return request

    def crawl(self, user, omit_status=(200,)):
        """
        Collect information to use in a more general test.
        """
        expect = {}
        for path in self.get_paths(user):
            request = self.get_processed_request(path)
            if request.response.get_status_code() not in omit_status:
                expect[path] = request.response.get_status_code()
        assert 0, 'expect=\n%s' % pformat(expect)


class PageTest(UTest):

    def __init__(self, site_name=None):
        loader = PageLoader(site_name=site_name)
        self.__class__._expect = loader.expect
        self.__class__._get_admin_user = loader.get_admin_user
        self.__class__._get_paths = loader.get_paths
        self.__class__._validate_html = loader.validate_html
        UTest.__init__(self)






"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_util.py $
$Id: utest_util.py 31303 2008-11-19 20:37:22Z dbinger $
"""
from datetime import datetime
from dulcinea.util import format_text, sanitize_url, wrap_paragraphs
from dulcinea.util import html2txt, activate_links, beginning_of_next_month
from dulcinea.util import _htmlescape_except, datetime_to_int, is_new, SAFE_TAGS
from dulcinea.util import insert_paragraph_tags, as_unicode, unichr
from dulcinea.util import UnicodeStringIO
from sancho.utest import UTest
from time import time
import sys

class UtilTest (UTest):

    def check_as_unicode_io(self):
        sio = UnicodeStringIO()
        sio.write('\u00a0')
        sio.write('\n')
        result = sio.getvalue()
        assert as_unicode(result)

    def check_html2txt(self):
        html = as_unicode('''
        <H1>Hello</H1>
        <i>11/17/04</i>
        <p>ido</p>
        <ul>
          <li> one
          <li> two
        </ul>
        <br>
        Bye now
        ''')
        text = as_unicode('\nHello\n\n11/17/04\n\nido\n\none\ntwo\n\nBye now')
        assert html2txt(html) == text, html2txt(html)
        assert html2txt(as_unicode(html)) == text
        html += as_unicode(' ') + unichr(0x99)
        text += as_unicode(' ') + unichr(0x99)
        html += as_unicode(' &Delta;')
        text += as_unicode(' ') + unichr(0x0394)
        assert html2txt(html) == text, (html2txt(html), text)

    def check_format_text(self):
        assert format_text('text') == 'text'
        format_text('http://www.example.org')

    def check_htmlescape_except(self):
        assert _htmlescape_except('<Br><bru>', SAFE_TAGS) == '<br>&lt;bru&gt;'

    def check_sanitize_url(self):
        assert sanitize_url(None) == None
        assert sanitize_url('www.example.org') == 'http://www.example.org'
        assert sanitize_url('nobody@example.com') == 'mailto:nobody@example.com'
        assert sanitize_url("ftp://ftp.example.com") == 'ftp://ftp.example.com'

    def check_wrap_paragraphs(self):
        assert wrap_paragraphs("text\n\nboo") == "text\n\nboo"
        assert wrap_paragraphs(None) == ''

    def check_insert_paragraph_tags(self):
        assert insert_paragraph_tags(None) == None
        assert str(insert_paragraph_tags('\n')) == '<p>\n</p>'

    def check_activate_links(self):
        assert activate_links(None) == None
        s = 'this http://example.org is fine'
        s_active = ('this <a href="http://example.org">http://example.org</a>'
                    ' is fine')
        assert str(activate_links(s)) == s_active
        s = 'this mailto:nobody@example.org is fine'
        s_active = (
            'this '
            'mailto:<a href="mailto:nobody@example.org">nobody@example.org</a>'
            ' is fine')
        assert str(activate_links(s)) == s_active
        s = 'this nobody@example.org is fine'
        s_active = (
            'this '
            '<a href="mailto:nobody@example.org">nobody@example.org</a>'
            ' is fine')
        assert str(activate_links(s)) == s_active
        s = 'this www.example.org/a. is fine'
        s_active = (
            'this '
            '<a href="http://www.example.org/a">www.example.org/a</a>.'
            ' is fine')
        assert str(activate_links(s)) == s_active

    def check_datetime_to_int(self):
        t = int(time())
        assert datetime_to_int(datetime.fromtimestamp(t)) == t
        for t in range(1, 10):
            assert datetime_to_int(datetime.fromtimestamp(t)) == t

    def check_beginning_of_next_month(self):
        a = datetime(year=2004, month=12, day=3, hour=1, minute=2, second=3,
                     microsecond=3)
        assert beginning_of_next_month(a).year == 2005
        assert beginning_of_next_month(a).month == 1
        assert beginning_of_next_month(a).day == 1
        assert beginning_of_next_month(a).hour == 0
        assert beginning_of_next_month(a).minute == 0
        assert beginning_of_next_month(a).second == 0
        assert beginning_of_next_month(a).microsecond == 0

    def check_is_new(self):
        class TestPersistent:
            def __init__(self, value):
                self._p_connection = value

        assert is_new(TestPersistent(None)) == True
        assert is_new(TestPersistent(1)) == False


if __name__ == "__main__":
    UtilTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/numeric.py $
$Id: numeric.py 31133 2008-09-17 02:26:31Z dbinger $
"""


def near_cmp(x, y, fudge = 2.5e-16):
    """Compare two numbers (presumably floating-point) with a fudge
    factor to account for accumulated round-off error.  Returns
    0, < 0, or > 0 just like the builtin 'cmp()'.
    """
    # The fudge factor for floating-point comparisons is 2.5e-16: this
    # is based on DBL_EPSILON from <float.h>, which is
    # 2.2204460492503131e-16 on every C compiler/hardware combination I
    # have access to right now (GCC/Intel, GCC/SPARC, GCC/MIPS, and
    # SGI/MIPS).  Must be an IEEE thing.
    if x is None and y is None:
        0
    elif x is None:
        return -1
    elif y is None:
        return 1
    elif isinstance(x, float) or isinstance(y, float):
        if x == 0.0 or y == 0.0:
            if abs(x - y) < fudge:
                return 0
        elif abs((x - y) / x) < fudge:
            return 0
    return cmp(x, y)





"""
$Id: note.py 29832 2007-05-29 19:19:08Z dbinger $
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/note.py $
"""
from dulcinea.attachable import Attachable
from durus.btree import BTree, BNode
from durus.persistent import PersistentObject
from qp.lib.spec import specify, string, sequence, add_getters_and_setters
from qp.lib.spec import init, Specified, Mixin, datetime_with_tz
from qp.pub.common import get_connection, get_user, site_now
from qp.pub.user import User

def get_note_mapping():
    return get_connection().get_root().get('note_mapping')

class Note (PersistentObject, Specified, Attachable):

    text_is = (string, None)
    timestamp_is = (datetime_with_tz, None)
    user_is = (User, None)
    title_is = (string, None)

    def __init__(self, text=None):
        Attachable.__init__(self)
        init(self, text=text)

    def set_timestamp(self, value=None):
        specify(self, timestamp=value or site_now())

    def set_user(self, user=None):
        specify(self, user=user or get_user())

    def is_empty(self):
        return not bool(self.text or self.get_attachments())

add_getters_and_setters(Note)

class Notable (Mixin):

    note_is = (Note, None)

    def __init__(self, note=None):
        self.set_note(note)

add_getters_and_setters(Notable)


class NoteMapping (BTree, Specified):

    root_is = sequence((string, Note), BNode)

    def add_note(self, note):
        title = note.get_title()
        if title.lower() not in self:
            self[title.lower()] = note

    def get_note(self, title, default=None):
        if title:
            return self.get(title.lower(), default=default)
        return default

    def remove_note(self, note):
        del self[note.get_title().lower()]




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/gallery.py $
$Id: gallery.py 29688 2007-03-29 13:46:15Z dbinger $
"""
from dulcinea.attachable import Attachable
from qp.lib.spec import Specified
from qp.pub.common import get_publisher
from durus.persistent import PersistentObject

def get_gallery():
    return get_publisher().get_connection().get_root().get('gallery')

class Gallery (PersistentObject, Specified, Attachable):

    def get_allowed_mime_types(self, user=None):
        return ['image/jpeg',
                'image/gif',
                'image/png']




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/user.py $
$Id: user.py 31139 2008-09-17 10:58:49Z dbinger $
"""
from durus.utils import as_bytes
from qp.lib.spec import add_getters_and_setters, specify
from qp.lib.spec import spec, pattern, Specified
from qp.lib.util import randbytes, sha1
from qp.pub.common import get_publisher, get_users
from qp.pub.user import User
import binascii


def hash_password(password):
    """Apply a one way hash function to a password and return the result."""
    return sha1(as_bytes(password)).hexdigest()

def get_realm():
    return get_publisher().get_site().get_name()

class DulcineaUser (User, Specified):
    """
    a registered user.
    """
    global_permissions = {
        "act-as":
            "Allow to act as another user.",
        "create-users":
            "Allow the creation of other users.",
        "manage-permissions":
            "Allow changing of permissions.",
        "staff":
            ("Is a member of the staff, with all of the privileges and "
             "responsibilities thereunto appertaining."),
        "system":
            "Allow to do things normally done by the software system.",
        }

    id_is = spec(
        pattern('^[-A-Za-z0-9_@. ]*$'),
        "unique among users here, spaces allowed")

    def __str__(self):
        return self.id or '*no id*'

    format = __str__ # subclasses should override

    def get_key(self):
        """ used for forming component representing this user in URLs
        """
        return self.get_id()

    def set_id(self, user_id):
        assert self.id == None
        specify(self, id=user_id)

    def set_password(self, new_password, realm=None, check=True):
        """Set the user's password to 'new_password'."""
        if check and self.check_new_password(new_password) != "":
            raise ValueError('invalid password')
        User.set_password(self, new_password, realm=realm)

    def has_password(self, password, realm=None):
        """Return true if the provided password is correct."""
        digester = self.get_digester()
        if realm is None:
            realm = get_realm()
        digest = digester.get_digest(realm)
        if User.has_password(self, password, realm=realm):
            return True
        elif digest == hash_password(password):
            # If the digest uses the older Dulcinea hash, set
            # a new digest using the QP hash.
            User.set_password(self, password, realm=realm)
            return True
        else:
            return False

    def generate_password (self, realm=None, length=6):
        """Set the password to a random value and return the new password."""
        password = binascii.b2a_base64(binascii.unhexlify(randbytes(length)))
        password = password[:length]
        self.set_password(password, realm=realm)
        return password

    def check_new_password(self, new_password):
        """(string) -> string
        Check if a new password is valid.  Returns the empty string if the
        password is okay otherwise returns a string that describes what is
        wrong with the entered password.
        """
        return ""

    def format_realname(self):
        return ''

    def is_null(self):
        return not self

    def is_disabled(self, realm=None):
        return not bool(self.get_digester().get_digest(realm or get_realm()))

    def is_system(self):
        return self.id == 'SYSTEM'

    def is_admin(self):
        return self.is_granted('staff')

    def is_granted(self, permission, granter=True):
        return self.get_permissions().is_granted(permission, granter)

    def can_manage_permissions(self):
        return self.is_granted('manage-permissions')

    def get_email_realname_pair(self):
        return self.get_email(), self.format_realname() or ''

add_getters_and_setters(DulcineaUser)

def list_users():
    return [user for user in get_users().itervalues()
           if not (user.is_null() or
                   user.is_system() or
                   user.is_disabled())]

def gen_users_granted(permission, granter=True):
    for user in get_users().itervalues():
        if user.is_granted(permission, granter):
            yield user

def get_matching_user(identifier):
    """(identifier : string) -> DulcineaUser | None

    Return a user with matching id or email address, or None
    if no such user is found.
    """
    return lookup_user(id=identifier, email=identifier)

def add_user(user):
    return get_publisher().add_user(user)

def lookup_user(id=None, email=None, **kwargs):
    return get_publisher().lookup_user(id=id, email=email, **kwargs)

def text_format_user(user):
    if not user:
        return 'None'
    realname = user.format_realname()
    if realname:
        text = '%s (%s)' % (realname, user.get_id())
    else:
        text = user.get_id()
    email = user.get_email()
    if email:
        text += ' <%s>' % email
    return text


"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/property/property_type.py $
$Id: property_type.py 31174 2008-09-24 17:59:59Z rmasse $
"""
from dulcinea.base import DulcineaPersistent
from dulcinea.material import Material, get_material_db
from dulcinea.physical_unit import get_standard_units
from dulcinea.physical_unit import PhysicalUnit, get_standard_unit
from dulcinea.physical_value import PhysicalValue
from dulcinea.property import get_property_db
from dulcinea.property.errors import PropertyTypeError, ConstraintError
from dulcinea.range_value import RangeValue
from dulcinea.set_utils import in_set
from durus.persistent import PersistentObject
from durus.persistent_dict import PersistentDict
from qp.lib.spec import mapping, anything, instance, spec, require, string
from qp.lib.spec import Specified
from qpy import xml, xml_quote, stringify
import dulcinea.property.property_template


class PropertyType (DulcineaPersistent):

    INT = 1
    STRING = 2
    PHYSICAL_VALUE = 3
    MATERIAL = 4
    BOOLEAN = 5

    ATOMIC_TYPES = ((INT, "integer"),
                    (STRING, "string"),
                    (PHYSICAL_VALUE, "physical value"),
                    (MATERIAL, "material"),
                    (BOOLEAN, 'boolean'))

    ATOMIC = 10
    LIST = 20
    TABLE = 30

    AGGREGATE_TYPES = ((ATOMIC, "atomic"),
                       (LIST, "list"),
                       (TABLE, "table"))

    def is_atomic (self):
        return 0

    def is_int (self):
        return 0

    def is_string (self):
        return 0

    def is_physical_value (self):
        return 0

    def is_material (self):
        return 0

    def is_boolean (self):
        return 0

    def is_list (self):
        return 0

    def is_table (self):
        return 0

    def is_discrete (self, constraint):
        raise NotImplementedError

    def allow_tolerance (self):
        return False

    def get_unit (self):
        return None

    def get_compatible_units (self):
        return None

    def check_value (self, value):
        """Checks that the *type* of value is compatible
        with this type of property.
        Raise PropertyTypeError if problems are discovered.
        """
        raise NotImplementedError

    def supports_simplify (self):
        return True

    def supports_constraints (self):
        "Returns true if setting constraints makes sense for this type."
        return True

    def check_constraint (self, constraint):
        """Checks that constraint is a sequence of
        items whose types are compatible with this type
        of property.
        Raise PropertyTypeError if problems are discovered.
        """
        if constraint is None:
            return
        if type(constraint) is list:
            for item in constraint:
                self.check_value(item)
        else:
            raise PropertyTypeError(
                'invalid constraint %r for %r' % (constraint, self))

    def check_value_in_constraint (self, value, constraint):
        if not in_set(value, constraint):
            self.value_in_constraint_error(value, constraint)

    def check_set_in_constraint (self, set, constraint):
        self.check_constraint(set)
        if constraint is None or set == constraint:
            return
        if set is None:
            raise ConstraintError
        for element in set:
            if not in_set(element, constraint):
                raise ConstraintError(
                    '%s not within constraint %s' % (element, constraint))


    def value_in_constraint_error (self, bad_value, constraint):
        raise ConstraintError(
            '%s not allowed: %s' % (bad_value,
                                    self.explain_constraint(constraint)))

    def explain_constraint (self, constraint, html=0):
        if html:
            texttype = xml
        else:
            texttype = str
        if not constraint:
            return None
        else:
            set = self.format_set(constraint, html=html)
            if set is None:
                return None
            elif len(set) <= 2:
                return texttype("must be ") + texttype(" or ").join(set)
            else:
                return texttype("must be one of ") + texttype(", ").join(set)

    def value_error (self, value):
        raise PropertyTypeError(
            'invalid value %r for %r' % (value, self))

    def format_value (self, value, html=0):
        if html:
            return xml_quote(value)
        else:
            return str(value)

    def format_set (self, set, html=False):
        """(set:[any], html:bool=False) -> [str|xml]
        Return a list of nice string (or xml) renderings of the elements
        of set, which is a constraint for this property type.

        Implemented in subclasses.
        """
        raise NotImplementedError

    def format_constraint (self, set, html=0):
        if set is not None:
            formatted_set = self.format_set(set, html=html)
            if formatted_set is not None:
                if html:
                    comma = xml(", ")
                else:
                    comma = ", "
                return comma.join(formatted_set)
        if html:
            return xml('<em>unconstrained</em>')
        else:
            return '(unconstrained)'


class IntPropertyType (PropertyType):

    def get_key (self):
        return (PropertyType.ATOMIC, PropertyType.INT,
                None, None, None)

    def is_atomic (self):
        return 1

    def is_int (self):
        return 1

    def is_discrete(self, constraint):
        if constraint is None:
            return False
        for item in constraint:
            if isinstance(item, RangeValue):
                return False
        return True

    def __str__ (self):
        return "Int"

    def check_value (self, value):
        if value is None:
            return
        if type(value) is int:
            return
        if isinstance(value, RangeValue):
            if type(value.get_min()) is int:
                return
        self.value_error(value)

    def simplify_value (self, value):
        return value

    def reconstitute_value (self, value):
        return value

    def format_set (self, set, html=0):
        if html:
            return list(map(xml_quote, set))
        else:
            return list(map(str, set))

    def get_constraint_range (self, constraint):
        if constraint is None:
            return None

        def get_min_and_max (val):
            if isinstance(val, RangeValue):
                return val.get_min(), val.get_max()
            else:
                return val, val

        val_min, val_max = get_min_and_max(constraint[0])
        for value in constraint[1:]:
            next_min, next_max = get_min_and_max(value)
            val_min = min(val_min, next_min)
            val_max = max(val_max, next_max)
        return RangeValue(val_min, val_max)

class StringPropertyType (PropertyType):

    def get_key (self):
        return (PropertyType.ATOMIC, PropertyType.STRING,
                None, None, None)

    def is_atomic (self):
        return 1

    def is_string (self):
        return 1

    def is_discrete (self, constraint):
        if constraint is None:
            return False
        else:
            return True

    def __str__ (self):
        return "String"

    def check_value (self, value):
        if value is None:
            return
        if string(value):
            return
        self.value_error(value)

    def simplify_value (self, value):
        return value

    def reconstitute_value (self, value):
        return value

    def format_set (self, set, html=0):
        return set

class BooleanPropertyType (PropertyType):


    def get_key (self):
        return (PropertyType.ATOMIC, PropertyType.BOOLEAN,
                None, None, None)

    def is_atomic (self):
        return 1

    def is_boolean (self):
        return 1

    def is_discrete(self, constraint):
        return True

    def __str__ (self):
        return "Boolean"

    def check_value (self, value):
        if value in (True, False, None):
            return
        self.value_error(value)

    def simplify_value (self, value):
        return value

    def reconstitute_value (self, value):
        return value

    def supports_constraints (self):
        return False

    def format_value (self, value, html=0):
        if value:
            return 'yes'
        else:
            return 'no'

    def format_constraint (self, set, html=0):
        if html:
            return xml('<em>yes</em> or <em>no</em>')
        else:
            return '(yes or no)'


class PhysicalValuePropertyType (PropertyType):

    sub_type_is = (None, PhysicalUnit)

    def __init__ (self, sub_type):
        assert sub_type is None or isinstance(sub_type, PhysicalUnit), (
            'invalid sub_type %r for %r' % (sub_type, self))
        self.sub_type = sub_type

    def get_key (self):
        return (PropertyType.ATOMIC, PropertyType.PHYSICAL_VALUE,
                None, self.sub_type, None)

    def get_unit (self):
        return self.sub_type

    def get_compatible_units (self):
        unit = self.get_unit()
        if not unit:
            return []
        return get_standard_units().get_compatible_units(unit)

    def is_atomic (self):
        return 1

    def is_physical_value (self):
        return 1

    def is_discrete(self, constraint):
        if constraint is None:
            return False
        for item in constraint:
            if item.is_range():
                return False
        return True

    def allow_tolerance(self):
        return True

    def __str__ (self):
        if self.sub_type is None:
            return "PhysicalValue"
        else:
            return "PhysicalValue (%s)" % self.sub_type

    def check_value (self, value):
        if value is None:
            return
        if isinstance(value, PhysicalValue):
            if self.sub_type is value.get_unit():
                return
            if self.sub_type and self.sub_type.is_compatible(value.get_unit()):
                return
        self.value_error(value)

    def get_constraint_range (self, constraint):
        if constraint is None:
            return None

        val_min = constraint[0].get_min()
        val_max = constraint[0].get_max()
        for value in constraint[1:]:
            val_min = min(val_min, value.get_min())
            val_max = max(val_max, value.get_max())
        assert val_min.get_unit() is val_max.get_unit(), (
            'unmatched units %r and %r for %r' % (val_min.get_unit(),
                                                  val_max.get_unit(),
                                                  self))
        return PhysicalValue(RangeValue(val_min.get_value(),
                                        val_max.get_value()),
                             val_min.get_unit())

    def simplify_value (self, value):
        default_unit = self.get_unit()
        if value.unit is not default_unit:
            value = value.convert(default_unit)
        return value.value

    def reconstitute_value (self, value):
        if string(value) and value.find(' .. ') != -1:
            value = RangeValue(*list(map(float, value.split(' .. ', 1))))
        return PhysicalValue(value, self.get_unit())

    def format_value (self, value, html=0):
        if isinstance(value, PhysicalValue):
            return value.format(html=html)
        return stringify(value)

    def format_set (self, set, html=0):
        items = [value.format(show_unit=0, html=html)
                 for value in set[0:-1]]
        items.append(set[-1].format(show_unit=1, html=html))
        return items


class MaterialPropertyType (PropertyType):

    def get_key (self):
        return (PropertyType.ATOMIC, PropertyType.MATERIAL,
                None, None, None)

    def is_atomic (self):
        return 1

    def is_material (self):
        return 1

    def is_discrete(self, constraint):
        return True

    def __str__ (self):
        return "Material"

    def check_value (self, value):
        if (value is None or
            isinstance(value, Material)):
            return
        self.value_error(value)

    def check_value_in_constraint (self, value, constraint):
        for material in constraint:
            if value.is_descendant_of(material):
                break
        else:
            self.value_in_constraint_error(value, constraint)

    def simplify_value (self, value):
        return value.get_name()

    def reconstitute_value (self, value):
        material = get_material_db().get_material(value)
        if material is None:
            raise PropertyTypeError('unknown material %r' % value)
        return material


    def format_set (self, set, html=0):
        if len(set) == len(get_material_db().get_materials()):
            return None
        else:
            return [value.get_label()
                    for value in set]


class ListPropertyType (PropertyType):

    element_type_is = PropertyType

    def __init__ (self, element_type):
        assert isinstance(element_type, PropertyType), (
            'invalid element_type %r for %r' % (element_type, self))
        self.element_type = element_type

    def get_key (self):
        element_key = self.element_type.get_key()
        return (PropertyType.LIST, element_key[1],
                None, element_key[3], None)

    def is_int (self):
        return self.element_type.is_int()

    def is_string (self):
        return self.element_type.is_string()

    def is_physical_value (self):
        return self.element_type.is_physical_value()

    def get_compatible_units (self):
        return self.element_type.get_compatible_units()

    def get_unit (self):
        return self.element_type.get_unit()

    def is_material (self):
        return self.element_type.is_material()

    def is_list (self):
        return 1

    def is_discrete(self, constraint):
        return self.element_type.is_discrete(constraint)

    def __str__ (self):
        return "List of %s" % self.element_type

    def check_value (self, value):
        if value is None:
            return
        if type(value) is list:
            for val in value:
                self.element_type.check_value(val)
            return
        self.value_error(value)

    def supports_simplify (self):
        return False

    def supports_constraints (self):
        return self.element_type.supports_constraints()

    def check_value_in_constraint (self, values, constraint):
        for value in values:
            self.element_type.check_value_in_constraint(value, constraint)

    def check_constraint (self, constraint):
        self.element_type.check_constraint(constraint)

    def format_value (self, value, html=0):
        fv = self.element_type.format_value
        values_format = [fv(val, html=html) for val in value]
        if html:
            comma = xml(", ")
        else:
            comma = ", "
        return comma.join(values_format)

    def format_set (self, set, html=0):
        return self.element_type.format_set(set, html=0)


class TablePropertyType (PropertyType):

    key_type_is = PropertyType
    element_type_is = PropertyType

    def __init__ (self, element_type, key_type):
        assert isinstance(element_type, PropertyType), (
            'invalid element_type %r for %r' % (element_type, self))
        assert isinstance(key_type, PropertyType), (
            'invalid key_type %r for %r' % (key_type, self))
        self.element_type = element_type
        self.key_type = key_type

    def get_key (self):
        element_key = self.element_type.get_key()
        key_key = self.key_type.get_key()
        return (PropertyType.TABLE, element_key[1],
                key_key[1], element_key[3], key_key[3])

    def is_int (self):
        return self.element_type.is_int()

    def is_string (self):
        return self.element_type.is_string()

    def is_physical_value (self):
        return self.element_type.is_physical_value()

    def get_compatible_units (self):
        return (self.key_type.get_compatible_units() or
                self.element_type.get_compatible_units())

    def get_unit (self):
        return self.element_type.get_unit()

    def is_material (self):
        return self.element_type.is_material()

    def is_key_int (self):
        return self.key_type.is_int()

    def is_key_string (self):
        return self.key_type.is_string()

    def is_key_physical_value (self):
        return self.key_type.is_physical_value()

    def is_key_material (self):
        return self.key_type.is_material()

    def is_table (self):
        return 1

    def is_discrete(self, constraint):
        assert constraint is None, 'cannot constrain %r' % self
        return False

    def get_key_type (self):
        return self.key_type

    def get_element_type (self):
        return self.element_type

    def __str__ (self):
        return "Table (%s to %s)" % (self.key_type, self.element_type)

    def check_value (self, value):
        if value is None:
            return
        if type(value) is dict:
            for key, val in value.items():
                self.key_type.check_value(key)
                self.element_type.check_value(val)
            return
        self.value_error(value)

    def supports_simplify (self):
        return False

    def supports_constraints (self):
        "Returns true if setting constraints makes sense for this type."
        return False

    def check_value_in_constraint (self, value, constraint):
        assert constraint is None, 'cannot constrain %r' % self

    def check_constraint (self, constraint):
        assert constraint is None, 'cannot constrain %r' % self

    def format_value (self, value, html=0):
        if html:
            texttype = xml
        else:
            texttype = str
        rows = [texttype("%s: %s") % (
                    self.key_type.format_value(key, html=html),
                    self.element_type.format_value(value, html=html))
                for key, value in value.items()]
        return texttype(", ").join(sorted(rows))


def create_atomic_type (element_type, element_sub_type=None):
    if element_type == PropertyType.INT:
        property_type = IntPropertyType()
    elif element_type == PropertyType.STRING:
        property_type = StringPropertyType()
    elif element_type == PropertyType.PHYSICAL_VALUE:
        property_type = PhysicalValuePropertyType(element_sub_type)
    elif element_type == PropertyType.MATERIAL:
        property_type = MaterialPropertyType()
    elif element_type == PropertyType.BOOLEAN:
        property_type = BooleanPropertyType()
    else:
        assert False, 'unknown atomic type %r' % element_type
    return property_type


def get_property_type (*args, **kw):
    return get_property_db().get_property_type(*args, **kw)

def get_physical_value_property_type (unit_string=None):
    unit = unit_string and get_standard_unit(unit_string)
    return get_property_type(PropertyType.ATOMIC,
                              PropertyType.PHYSICAL_VALUE,
                              element_sub_type=unit)

def get_list_of_physical_value_property_type (unit_string=None):
    unit = unit_string and get_standard_unit(unit_string)
    return get_property_type(PropertyType.LIST,
                              PropertyType.PHYSICAL_VALUE,
                              element_sub_type=unit)

def get_int_property_type ():
    return get_property_type(PropertyType.ATOMIC,
                              PropertyType.INT)

def get_list_of_int_property_type ():
    return get_property_type(PropertyType.LIST,
                              PropertyType.INT)

def get_string_property_type ():
    return get_property_type(PropertyType.ATOMIC,
                              PropertyType.STRING)

def get_list_of_string_property_type ():
    return get_property_type(PropertyType.LIST,
                              PropertyType.STRING)

def get_boolean_property_type ():
    return get_property_type(PropertyType.ATOMIC,
                              PropertyType.BOOLEAN)

def get_material_property_type ():
    return get_property_type(PropertyType.ATOMIC,
                              PropertyType.MATERIAL)

def get_list_of_material_property_type ():
    return get_property_type(PropertyType.LIST,
                              PropertyType.MATERIAL)

def get_string_int_table_property_type ():
    return get_property_type(PropertyType.TABLE,
                              PropertyType.INT,
                              key_type=PropertyType.STRING)

def get_physical_value_int_table_property_type (unit_string=None):
    unit = unit_string and get_standard_unit(unit_string)
    return get_property_type(PropertyType.TABLE,
                              element_type=PropertyType.INT,
                              key_type=PropertyType.PHYSICAL_VALUE,
                              key_sub_type=unit)



class PropertyDatabase (PersistentObject, Specified):

    property_types_is = spec(
        mapping({(int, int, (None, int), anything, anything): PropertyType},
                PersistentDict),
        "The first three elements of the tuple are constants, defined "
        "in PropertyType that identify the aggregate type, the key "
        "type, and the element type of this particular PropertyType. "
        "The last two elements of the tuple identify subtypes of "
        "the key and the element types.  These subtypes are "
        "usually instances of PhysicalUnit.")
    master_templates_is = spec(
        mapping({string:instance('MasterTemplate')}, PersistentDict),
        "master templates by name")

    def __init__(self):
        self.property_types = PersistentDict()
        self.master_templates = PersistentDict()


    def add_master_template(self, master):
        require(master, dulcinea.property.property_template.MasterTemplate)
        if self.master_templates.has_key(master.get_name()):
            raise ValueError(
                "MasterTemplate '%s' is already defined." % master.get_name())
        self.master_templates[master.get_name()] = master

    def rename_master_template(self, master, new_name):
        """(master : MasterTemplate, new_name : string)

        This should only be called by MasterTemplate.set_name()
        """
        if not self.master_templates.has_key(master.get_name()):
            raise ValueError(
                "MasterTemplate '%s' is already defined." % master.get_name())
        if self.master_templates.has_key(new_name):
            raise ValueError(
                "MasterTemplate '%s' is already defined." % master.get_name())
        old_name = master.get_name()
        master.set_name(new_name)
        del self.master_templates[old_name]
        self.master_templates[new_name] = master

    def get_master_template(self, name):
        return self.master_templates.get(name)

    def get_master_templates(self):
        return list(self.master_templates.values())

    def _create_property_type(self,
                               aggregate_type,
                               element_type,
                               key_type=None,
                               element_sub_type=None,
                               key_sub_type=None):
        assert aggregate_type is PropertyType.TABLE or (
            key_type is None and key_sub_type is None), (
            "invalid type creation: key_type %r or key_sub_type %r "
            "defined for non table type" % (key_type, key_sub_type))
        assert element_type is PropertyType.PHYSICAL_VALUE or (
            element_sub_type is None), (
            "invalid type creation: element_type %r does not take "
            "sub_type argument %r" % (element_type, element_sub_type))

        if aggregate_type is PropertyType.ATOMIC:
            property_type = create_atomic_type(element_type,
                                                element_sub_type)

        elif aggregate_type is PropertyType.LIST:
            element_inst = self.get_property_type(
                PropertyType.ATOMIC, element_type,
                element_sub_type=element_sub_type)
            property_type = ListPropertyType(element_inst)

        elif aggregate_type is PropertyType.TABLE:
            assert key_type is PropertyType.PHYSICAL_VALUE or (
                key_sub_type is None), (
                "invalid type creation: key_type %r does not take "
                "sub_type argument %r" % (key_type, key_sub_type))
            element_inst = self.get_property_type(
                PropertyType.ATOMIC, element_type,
                element_sub_type=element_sub_type)
            key_inst = self.get_property_type(
                PropertyType.ATOMIC, key_type, element_sub_type=key_sub_type)
            property_type = TablePropertyType(element_inst, key_inst)

        else:
            assert False, (
                "unknown aggregate type: %r" % aggregate_type)

        self.property_types[(aggregate_type,
                              element_type,
                              key_type,
                              element_sub_type,
                              key_sub_type)] = property_type

        return property_type


    def get_property_type(self,
                           aggregate_type,
                           element_type,
                           key_type=None,
                           element_sub_type=None,
                           key_sub_type=None):
        property_type = self.property_types.get((aggregate_type,
                                                   element_type,
                                                   key_type,
                                                   element_sub_type,
                                                   key_sub_type))
        if property_type is None:
            property_type = self._create_property_type(aggregate_type,
                                                         element_type,
                                                         key_type,
                                                         element_sub_type,
                                                         key_sub_type)
        return property_type




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/child_process.py $
$Id: child_process.py 30050 2007-09-05 15:04:09Z dbinger $

Execute a child process and communicate with it using pipes.  This interface
has the following advantages over os.system() and os.popen():

  * The shell is not used for parsing the command line.  This eliminates
    problems with shell meta characters.

  * Pipes to stdin, stdout and stderr are provided.

  * The exit status of the child process is available.
     
"""
import os
import sys

class _ChildProcess:
    """
    Instance attributes:
    
        pid : int
          the process id of the child process
        stdin : file
          open file connected to file descriptor 0 of the child process
        stdout : file
          open file connected to file descriptor 1 of the child process
        stderr : file
          open file connected to file descriptor 2 of the child process
    """

    try:
        MAXFD = os.sysconf('SC_OPEN_MAX')
    except:
        MAXFD = 256

    def __init__(self, argv, bufsize=-1):
        child_stdin, stdin = os.pipe()
        stdout, child_stdout = os.pipe()
        stderr, child_stderr = os.pipe()
        self.stdin = os.fdopen(stdin, 'w', bufsize)
        self.stdout = os.fdopen(stdout, 'r', bufsize)
        self.stderr = os.fdopen(stderr, 'r', bufsize)
        self.pid = os.fork()
        if self.pid == 0:
            exec_error_fd = sys.stderr.fileno()
            # ensure exec_error_fd and friends are not using fds 0-3
            while exec_error_fd < 3:
                exec_error_fd = os.dup(exec_error_fd)
            while child_stdout < 3:
                child_stdout = os.dup(child_stdout)
            while child_stderr < 3:
                child_stderr = os.dup(child_stderr)
            os.dup2(child_stdin, 0)
            os.dup2(child_stdout, 1)
            os.dup2(child_stderr, 2)
            os.dup2(exec_error_fd, 3)
            for i in range(4, self.MAXFD):
                try:
                    os.close(i)
                except:
                    pass
            try:
                os.execvp(argv[0], argv)
            except OSError:
                err = sys.exc_info()[1]
                os.write(3, "couldn't exec %s: %s\n" % (argv[0], err.strerror))
            except:
                os.write(3, "couldn't exec %s\n" % argv[0])
            os._exit(1)
        os.close(child_stdin)
        os.close(child_stdout)
        os.close(child_stderr)

    def read(self, n=-1):
        """Read from child stdout"""
        return self.stdout.read(n)

    def readline(self):
        """Readline from child stdout"""
        return self.stdout.readline()

    def readlines(self):
        """Readlines from child stdout"""
        return self.stdout.readlines()

    def write(self, s):
        """Write data to child stdin"""
        self.stdin.write(s)

    def flush(self):
        """Flush stdin pipe to child"""
        self.stdin.flush()

    def close(self):
        """Close stdin, stdout and stderr pipes to child process.  Wait
        for the exit status of the child and return it."""
        for fd in (self.stdin, self.stdout, self.stderr):
            if not fd.closed:
                fd.close()
        status = self.wait()
        if status == 0:
            return None # matches behavior of popen(...).close()
        else:
            return status

    def wait(self, flags=0):
        pid, status = os.waitpid(self.pid, flags)
        return status


def execute(argv, bufsize=-1):
    """(argv : (string*), bufsize=-1) -> _ChildProcess
    
    Create a child process with pipes connected to its stdout, stdin and
    stderr file descriptors.  The child is created using fork() and
    execvp() in order to avoid the argument quoting problems presented by
    using system() or popen().  The bufsize argument has the same meaning as
    for the open() builtin and applies to stdin, stdout, and stderr.

    Examples:

        >>> p = execute(("ls", "/"))
        >>> p.readline()
        'bin\n'
        >>> p.readline()
        'boot\n'
        >>> p.close()
        >>>

        >>> p = execute(("cat",))
        >>> p.write("hello world\n")
        >>> p.stdin.close()
        >>> p.read()
        'hello world\n'
        >>> p.close()
        >>>

        >>> p = execute(("false",))
        >>> p.close()
        256
        >>>
    """
    return _ChildProcess(argv, bufsize)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/life_cycle_util.py $
$Id: life_cycle_util.py 31285 2008-11-16 23:11:32Z dbinger $
"""
from dulcinea.metaclass import ClassMethodsClass
from dulcinea.sort import lexical_sort
from qp.lib.spec import Mixin, callable
import sys

class TransitionClass (ClassMethodsClass):
    """This class is here so that we have a distinctive type for Transitions.

    Instance attributes: none
    """

    def __init__ (self, class_name, bases, namespace):
        super(TransitionClass, self).__init__(class_name, bases, namespace)
        # provide default values for attributes
        for attr in ['_progressive', '_require_comment', '_suppress_comment']:
            if not hasattr(self, attr):
                setattr(self, attr, 0)

Transition = TransitionClass('Transition', (object,), {})

class StateClass (ClassMethodsClass):
    """This class is here so that we have a distinctive type for States.

    Instance attributes:
      _name : string
      _transitions : [TransitionClass]
    """

    def __init__ (self, class_name, bases, namespace):
        super(StateClass, self).__init__(class_name, bases, namespace)
        # compute _transitions list
        self._transitions = lexical_sort(
            [ value for value in namespace.values()
              if isinstance(value, TransitionClass) ])

State = StateClass('State', (object,), {})


class PlainState (State):

    """
    Subclasses of PlainState are states of Statefuls. !

    Instance attributes: none
    """

    _name = "PlainState"
    _transitions = []

    def __init__ (self):
        """
        No instances of PlainState should ever exist.
        Subclasses are used directly as if they were instances.
        """
        raise RuntimeError

    def get_name (klass):
        """() -> string
        A pretty label for the state.
        """
        return klass._name

    def get_transitions (klass):
        """() -> sequence [ Transition ]
        Generates the list of Transitions in this class.
        """
        return klass._transitions

    def get_transition (klass, action):
        """() -> Transition
        Return the Transition with the given action,
        or None if no such transition is present.
        """
        for transition in klass.get_transitions():
            if action == transition.get_action():
                return transition
        return None

    def validate (*klasses):
        """
        Look for coding mistakes in this class.
        """
        for klass in klasses:
            assert type(klass.get_name()) is str
            actions = []
            for transition in klass.get_transitions():
                action = transition.get_action()
                assert action not in actions
                actions.append(action)
                transition.validate()

    def format (klass):
        """() -> string
        Return a plain text dump of this class.
        """
        s = 'State %s\n' % klass
        s += '    name=%r\n' % klass.get_name()
        s += '    '
        for transition in klass.get_transitions():
            s += transition.format().replace('\n','\n    ')
        s += '\n'
        return  s


class PlainTransition (Transition):
    """
    Subclasses of this class are defined within subclasses of State.

    Instance attributes: none
    """

    _action = _description = _hint = ''
    _next_state = PlainState

    def __init__ (self):
        """
        No instances of PlainTransition should ever exist.
        Subclasses are used directly as if they were instances.
        """
        raise RuntimeError

    def get_action (klass):
        """() -> string

        Return the string that is used to identify this transition.
        """
        return klass._action

    def get_description (klass):
        """() -> string
        """
        return klass._description

    def get_hint (klass):
        """() -> string

        A hint to users about this transition.
        """
        return klass._hint

    def get_next_state (klass):
        """() -> State
        Return the state that the run should be in after this transition
        is executed.
        """
        return klass._next_state

    def disallowed (klass, stateful, user):
        """(stateful : Stateful, user : User) ->  string | None
        Return an explanation of any problem (that the user can't correct)
        that prevents the user from executing this transition for the stateful.
        If this returns a string, this transition should not be offered.

        Subclasses will override this method.
        """
        return ("The %s.%s transition isn't allowed for anyone." %
                (stateful.state, klass))

    def get_warnings (klass, stateful, user, comment):
        """(stateful : Stateful, user : User) -> [ string ]
        Return a list of explanations of any problems (that the user could
        possibly correct) that prevent the user from executing this transition
        for the stateful.  If this returns any strings, this transition may be
        shown in some disabled form to the user, along with the explanation of
        why it is not currently available.

        Subclasses will override this method.
        """
        return []

    def prepare (klass, stateful, user, comment):
        """(stateful : Stateful, user : User, comment : string=None)
            -> string | None
        Assuming that disallowed() and get_warnings() return no reasons
        to stop, the transition may be attempted.  Sometimes there is something
        destructive that should be done just prior to the actual change
        of state, and the change may not produce the conditions that
        are required for the actual change of state to occur.
        This method executes those destructive operations and
        reports any final reason why the actual state change
        should not happen.

        Subclasses will override this method.
        """
        return None

    def change_state (klass, stateful, user, comment):
        """(stateful : Stateful, user : User, comment : string)
            -> [ string ] | None
        Try to execute this transition.
        Make the necessary history records,
        send the necessary email,
        perform the necessary data modifications,
        and assign a State to the stateful's 'state' attribute.

        If there are some reasons that the transition can't be completed,
        return a string containing the reasons, otherwise return None.
        """
        errors  = (klass.disallowed(stateful, user) or
                   klass.get_warnings(stateful, user, comment) or
                   klass.prepare(stateful, user, comment))
        if errors:
            if type(errors) is str:
                errors = [ errors ]
            klass.mail_on_errors(stateful, user, comment, errors)
            return errors
        else:
            stateful.set_state(klass.get_next_state(), user)
            klass.mail_on_success(stateful, user, comment)
            return None

    def is_progressive (klass, stateful=None, user=None):
        """(stateful : Stateful, user : User) -> boolean
        Is this a transition that moves the stateful toward completion?
        """
        return klass._progressive

    def require_comment (klass, stateful=None, user=None):
        """(stateful : Stateful, user : User) -> boolean
        Should the user be required to provide a comment when executing
        this transition?
        """
        return klass._require_comment

    def suppress_comment (klass, stateful=None, user=None):
        """(stateful : Stateful, user : User) -> boolean
        Should the user be offered a chance to comment when executing
        this transition?
        """
        return klass._suppress_comment

    def mail_on_errors (klass, stateful, user, comment, errors):
        """(stateful:Stateful, user:User, comment:string, errors:[string])
            -> (non_fab_message : string, fab_message : string)
        Deliver notification about this failed transition to the
        appropriate parties.  Subclasses will override this
        if they want any require any notification.

        Return a pair of strings that report the messages
        sent.  The first string describes the message that would
        be sent to all except fab staff.  The second string
        describes the message that would be sent to fab staff.
        The difference exists because we don't identify the
        customer to fab staff.
        """
        return ('No mail\n', '')

    def mail_on_success (klass, stateful, user, comment):
        """(stateful : Stateful, user : User, comment : string)
            -> (non_fab_message : string, fab_message : string)
        Deliver notification about this successful transition to the
        appropriate parties.  Subclasses will override this
        if they want any require any notification.

        Return a pair of strings that report the messages
        sent.  The first string describes the message that would
        be sent to all except fab staff.  The second string
        describes the message that would be sent to fab staff.
        The difference exists because we don't identify the
        customer to fab staff.
        """
        return ('No mail\n', '')

    def validate (*klasses):
        """
        Look for coding mistakes in this class.
        """
        for klass in klasses:
            assert issubclass(klass, PlainTransition)
            assert type(klass.get_action()) is str
            assert type(klass.get_description()) is str
            assert type(klass.get_hint()) is str
            assert issubclass(klass.get_next_state(), PlainState)
            assert callable(klass.require_comment)
            assert callable(klass.suppress_comment)
            assert callable(klass.is_progressive)
            assert callable(klass.disallowed)
            assert callable(klass.get_warnings)
            assert callable(klass.prepare)
            assert callable(klass.mail_on_errors)
            assert callable(klass.mail_on_success)
            assert callable(klass.change_state)

    def format (klass):
        """() -> string

        Return a plain text dump of this class.
        """
        s =  'Transition %s (%s) to State %s' % (klass,
                                                 klass.get_action(),
                                                 klass.get_next_state())
        s += '\n'
        s += '    description=%r\n' % klass.get_description()
        s += '    hint=%r\n' % klass.get_hint()
        s += '    is_progressive=%s\n' % klass.is_progressive()
        s += '    disallowed=%s\n' % klass.disallowed.func_name
        s += '    require_comment=%s\n' % klass.require_comment()
        s += '    suppress_comment=%s\n' % klass.suppress_comment()
        return s

class Stateful (Mixin):

    """Stateful objects have a state attribute.

    Instance attributes:
       state : StateClass
    """
    state_is = StateClass

    def __init__ (self, state):
        self.state = state

    def is_ready_for_user (self, user):
        """(user : User) -> boolean

        Return true if user is able to move the run forward in its life cycle.
        """
        for transition in self.state.get_transitions():
            if (transition.is_progressive(self, user) and
                transition.disallowed(self, user) is None):
                return 1
        return 0

    def get_available_actions (self, user):
        """(user:User) -> [(action,description,hint)]
        Returns a list of the actions that can be successfully taken
        by the given user.
        """
        return [ (transition.get_action(),
                  transition.get_description(),
                  transition.get_hint())
                 for transition in self.state.get_transitions()
                 if transition.disallowed(self, user) is None ]

    def get_transition (self, user, action):
        """(user : User, action : string) -> PlainTransition

        Look up the FSM transition for this user and action.  Return None if
        the transition does not exist or if it is not valid for this user.
        """
        transition = self.state.get_transition(action)
        if transition is None or transition.disallowed(self, user):
            return None
        else:
            return transition

    def change_state (self, user, action, comment=None):
        """(user : User, action : string, comment : string) -> [ string ]

        Change the run's state depending on the provided acting user
        and action.  If anything prevents the transition from happening,
        return a list of reasons.
        """
        transition = self.state.get_transition(action)
        if transition is None:
            return [ "No %s action available from %s."
                     % (action, self.get_state()) ]
        return transition.change_state(self, user, comment)

    def set_state (self, state, user):
        self.state = state

    def get_state (self):
        return self.state

def find_states (namespace):
    """
    Return a sorted list of states defined in the namespace
    and update the _next_state attributes.

    A module that defines a life cycle should call this after
    defining states, with a line, for example, like this:

    states = find_states(globals())

    """
    state_dict = {}
    for name, value in namespace.items():
        if isinstance(value, StateClass):
            state_dict[name] = value
    # connect transitions to states
    for state in state_dict.values():
        for transition in state.get_transitions():
            transition._next_state = state_dict[transition._next_state_name]
    return lexical_sort(state_dict.values())







"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/set_utils.py $
$Id: set_utils.py 30314 2007-11-27 12:15:22Z dbinger $

Provides utility functions for performing set operations with range values,
such as intersection and union.
"""
from dulcinea.physical_value import PhysicalValue
from dulcinea.range_value import RangeValue
from qp.lib.spec import require, either, number_classes

def in_set(val, set):
    """(val : any, set : list|tuple) -> boolean

    Return true if 'val' is in 'set'. If set contains any ranges it will
    return true if 'val' is within one of those ranges.
    """
    require(set, either(list, tuple))
    for item in set:
        if val == item:
            return 1
        elif (isinstance(item, (RangeValue, PhysicalValue)) and
              item.in_range(val)):
            return 1
    return 0

def get_range_intersection(range1, range2):
    """( RangeValue | int | long | float,
         RangeValue | int | long | float ) ->
         None | RangeValue | int | long | float

    Returns the intersection.
    """
    def lo_hi(x):
        if isinstance(x, RangeValue):
            return x.get_min(), x.get_max()
        else:
            require(x, either(*number_classes))
            return x, x

    lo1, hi1 = lo_hi(range1)
    lo2, hi2 = lo_hi(range2)

    lo = max(lo1, lo2)
    hi = min(hi1, hi2)

    if lo > hi:
        return None
    elif lo == hi:
        return lo
    else:
        return RangeValue(lo, hi)


def _get_physical_value_intersection(physical_value1, physical_value2):
    require(physical_value1, PhysicalValue)
    require(physical_value2, PhysicalValue)
    range2 = physical_value1.get_comparable_value(physical_value2)
    if range2 is None:
        return None
    intersection = get_range_intersection(physical_value1.get_value(), range2)
    if intersection is None:
        return None
    else:
        return PhysicalValue(intersection, physical_value1.get_unit())

def get_set_intersection(set1, set2):
    """( list , list ) -> list

    Returns the intersection of two sets. For sets containing ranges, all of
    the items in the two sets must be either numbers (RangeValue, int, long,
    float) or PhysicalValues.
    """
    require(set1, list)
    require(set2, list)
    set = []
    for item1 in set1:
        for item2 in set2:
            if item1 == item2:
                intersection = item1
            elif isinstance(item1, PhysicalValue):
                intersection = _get_physical_value_intersection(item1, item2)
            elif (isinstance(item1, RangeValue) or
                  isinstance(item2, RangeValue)):
                intersection = get_range_intersection(item1, item2)
            if intersection is not None:
                set.append(intersection)
    return set




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/timestamped.py $
$Id: timestamped.py 29840 2007-05-30 16:00:09Z rmasse $
"""
from dulcinea.sort import attr_sort
from qp.lib.spec import Mixin, datetime_with_tz
from qp.pub.common import site_now

def timestamp_sorted(timestamped_sequence):
    return attr_sort(timestamped_sequence, 'timestamp')

def reverse_timestamp_sorted(timestamped_sequence):
    result = timestamp_sorted(timestamped_sequence)
    result.reverse()
    return result

class Timestamped (Mixin):

    timestamp_is = datetime_with_tz

    def __init__(self):
        self.set_timestamp()

    def get_timestamp(self):
        return self.timestamp

    def set_timestamp(self):
        self.timestamp = site_now()


def convert_datetimes_to_timezone_aware(connection, default_tzinfo=None):
    """
    This is for use in database update scripts.
    """
    from qp.lib.tz import Eastern
    import datetime as datetime_module
    old_datetime = datetime_module.datetime
    tz = default_tzinfo or Eastern
    class datetime (old_datetime):
        number_of_datetimes = 0
        default_tzinfo = tz
        def __new__(klass, *args, **kwargs):
            datetime.number_of_datetimes += 1
            self = old_datetime.__new__(old_datetime, *args, **kwargs)
            if self.tzinfo is None:
                self = self.replace(tzinfo=klass.default_tzinfo)
            assert self.__class__ is old_datetime
            return self
    datetime_module.datetime = datetime
    datetime.__module__ = 'datetime'
    previous = datetime.number_of_datetimes
    for object in connection.get_crawler():
        if datetime.number_of_datetimes > previous:
            object._p_note_change()
            previous = datetime.number_of_datetimes
    datetime_module.datetime = old_datetime



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/material.py $
$Id: material.py 30544 2008-03-07 20:43:18Z dbinger $
"""
from dulcinea.category import Category
from durus.persistent import PersistentObject
from durus.persistent_dict import PersistentDict
from durus.persistent_list import PersistentList
from qp.lib.spec import boolean, sequence, instance, add_getters_and_setters
from qp.lib.spec import spec, specify, require, string, Specified, mapping
from qp.pub.common import get_publisher
from qpy import xml, xml_quote
from random import randrange


def get_material_db():
    return get_publisher().get_connection().get_root().get('material_db')

class Material (Category):
    """
    Represents a material and a category of materials.
    """
    html_name_is = spec(
        (string, None),
        "The name to be used in html displays. "
        "This may include html tags for subscripts. "
        "Other materials may have the same html_name, "
        "but that is probably a bad idea.")
    html_color_is = spec(
        (string, None),
        "A string in the form #ffee99 that identifies "
        "a color that might be used in html concerning "
        "this Material. "
        "Other materials may have the same color.")
    deleted_is = spec(
        boolean,
        "Tells if this this material is deprecated.")

    children_is = sequence(instance('Material'), PersistentList)
    parents_is = sequence(instance('Material'), PersistentList)

    def __init__(self, name, html_name=None, html_color=None):
        Category.__init__(self)
        specify(self,
                name=name,
                html_color=html_color or random_light_color(),
                html_name=html_name or name,
                deleted=0)

    def __str__(self):
        if self.get_children():
            return '%s (category)' % self.get_label()
        else:
            return self.get_label()

    def get_label(self):
        return self.label or self.name

    def format(self, html=False):
        if html:
            if self.html_name:
                return xml(self.html_name)
            else:
                return xml_quote(self.get_label())
        else:
            return self.get_label()

    def is_deleted(self):
        return self.deleted

    def is_other(self):
        return self.name == "other"

add_getters_and_setters(Material)

def random_light_color():
    red   = 256
    green = 256
    blue  = 256
    maximum_light = (256*3-100)
    minimum_light = 300
    while ((red+green+blue > maximum_light) or
           (red+green+blue < minimum_light)):
        red = randrange(16,256)
        green = randrange(16,256)
        blue = randrange(16,256)
    r = hex(red)[-2:]
    g = hex(green)[-2:]
    b = hex(blue)[-2:]
    return '#%s%s%s' % (r,g,b)

def material_sort(materials):
    """(seq [Material]) -> [Material]"""
    def lower_case_label(material):
        return material.get_label().lower()
    return sorted(materials, key=lower_case_label)


class MaterialDatabase (PersistentObject, Specified):

    materials_is = spec(
        mapping({string:Material}, PersistentDict),
        "all material objects keyed on name")

    def __init__(self):
        specify(self, materials=PersistentDict())

    def __getitem__(self, material_name):
        return self.materials[material_name]

    def add_material(self, material):
        require(material, Material)
        if self.materials.has_key(material.get_name()):
            raise ValueError(
                "material named '%s' already exists" % material.get_name())
        self.materials[material.get_name()] = material

    def get_material(self, material_name):
        return self.materials.get(material_name)

    def get_materials(self):
        return list(self.materials.values())




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/base.py $
$Id: base.py 31344 2008-11-21 20:45:40Z dbinger $
"""
from durus.persistent import PersistentObject
from copy import copy
from qp.lib.spec import Specified, Mixin

class Copyable (Mixin):

    def copy(self):
        return copy(self)


class DulcineaPersistent (PersistentObject, Specified, Copyable):


    def __copy__(self):
        new_instance = PersistentObject.__new__(self.__class__)
        new_instance.__setstate__(self.__getstate__())
        return new_instance

    def __eq__(self, other):
        return (self.__class__ == getattr(other, '__class__', None) and
            other.__getstate__() == self.__getstate__())

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return id(self)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/property/test/utest_property_type.py $
$Id: utest_property_type.py 28932 2006-10-12 20:29:05Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.material import get_material_db
from dulcinea.property import get_property_db
from dulcinea.property.errors import PropertyTypeError
from dulcinea.property.property_type import PropertyType
from dulcinea.physical_unit import get_standard_unit
from dulcinea.physical_value import PhysicalValue
from dulcinea.range_value import RangeValue
from sancho.utest import raises

class PropertyTypeTest (CommonTest):

    def _pre (self):
        CommonTest._pre(self)
        if get_property_db() is None:
            return
        self.plib = get_property_db()
        self.mdb = get_material_db()

    def check_int_type (self):
        if get_property_db() is None:
            return
        property_type1 = self.plib.get_property_type(PropertyType.ATOMIC,
                                                       PropertyType.INT)

        assert property_type1.is_atomic()
        assert property_type1.is_int()
        assert not property_type1.is_string()
        assert not property_type1.is_physical_value()
        assert not property_type1.is_material()
        assert not property_type1.is_list()
        assert not property_type1.is_table()
        assert not property_type1.is_boolean()

        property_type2 = self.plib.get_property_type(PropertyType.ATOMIC,
                                                       PropertyType.STRING)
        assert property_type2 is not property_type1
        property_type2 = self.plib.get_property_type(PropertyType.ATOMIC,
                                                       PropertyType.INT)
        assert property_type2 is property_type1

        property_type1.check_value(None)
        property_type1.check_value(10)
        property_type1.check_value(RangeValue(10, 20))
        raises(PropertyTypeError, property_type1.check_value, [10, 20])
        raises(PropertyTypeError, property_type1.check_value, 'hello')
        raises(PropertyTypeError, property_type1.check_value, ['hello', 'bye'])
        raises(AssertionError,
            self.plib.get_property_type,
            PropertyType.ATOMIC, PropertyType.INT, key_type=1)

    def check_string_type (self):
        if get_property_db() is None:
            return
        property_type1 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.STRING)

        assert property_type1.is_atomic()
        assert not property_type1.is_int()
        assert property_type1.is_string()
        assert not property_type1.is_physical_value()
        assert not property_type1.is_material()
        assert not property_type1.is_list()
        assert not property_type1.is_table()
        assert not property_type1.is_boolean()

        property_type2 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.INT)
        assert property_type2 is not property_type1
        property_type2 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.STRING)
        assert property_type2 is property_type1

        property_type1.check_value(None)
        property_type1.check_value('hello')
        raises(PropertyTypeError, property_type1.check_value, ['hello', 'bye'])
        raises(PropertyTypeError, property_type1.check_value, 10)
        raises(PropertyTypeError, property_type1.check_value, RangeValue(10, 20))
        raises(PropertyTypeError, property_type1.check_value, [10, 20])

    def check_boolean_type (self):
        if get_property_db() is None:
            return
        property_type1 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.BOOLEAN)

        assert property_type1.is_atomic()
        assert not property_type1.is_int()
        assert not property_type1.is_string()
        assert not property_type1.is_physical_value()
        assert not property_type1.is_material()
        assert not property_type1.is_list()
        assert not property_type1.is_table()
        assert property_type1.is_boolean()

        property_type2 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.INT)
        assert property_type2 is not property_type1
        property_type2 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.BOOLEAN)
        assert property_type2 is property_type1

        property_type1.check_value(True)
        property_type1.check_value(False)
        property_type1.check_value(0)
        property_type1.check_value(1)
        property_type1.check_value(None)
        raises(PropertyTypeError, property_type1.check_value, '')
        raises(PropertyTypeError, property_type1.check_value, [True])
        assert property_type1.format_value(True) == 'yes'
        assert property_type1.format_value(0) == 'no'

    def check_physical_value_type (self):
        if get_property_db() is None:
            return
        property_type1 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.PHYSICAL_VALUE)

        assert property_type1.is_atomic()
        assert not property_type1.is_int()
        assert not property_type1.is_string()
        assert property_type1.is_physical_value()
        assert not property_type1.is_material()
        assert not property_type1.is_list()
        assert not property_type1.is_table()
        assert not property_type1.is_boolean()

        property_type2 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.PHYSICAL_VALUE)
        assert property_type2 is property_type1

        inch = get_standard_unit('inch')
        property_type2 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.PHYSICAL_VALUE,
            element_sub_type=inch)
        assert property_type2 is not property_type1

        property_type1.check_value(None)
        property_type1.check_value(PhysicalValue(10))
        property_type1.check_value(PhysicalValue(RangeValue(10, 20)))
        property_type2.check_value(PhysicalValue(10, 'inch'))
        property_type2.check_value(PhysicalValue(RangeValue(10, 20), 'inch'))
        property_type2.check_value(PhysicalValue(10, 'm'))
        property_type2.check_value(PhysicalValue(RangeValue(10, 20), 'm'))
        raises(PropertyTypeError, property_type1.check_value, 
            PhysicalValue(10, 'inch'))
        raises(PropertyTypeError, property_type2.check_value, PhysicalValue(10))
        raises(PropertyTypeError, property_type2.check_value, 
            PhysicalValue(10, 's'))
        raises(PropertyTypeError, property_type1.check_value, [PhysicalValue(10)])
        raises(PropertyTypeError, property_type2.check_value,
            [PhysicalValue(10, 'inch')])
        raises(PropertyTypeError,
            property_type1.check_value, 10)
        raises(PropertyTypeError,
            property_type1.check_value, RangeValue(10, 20))
        raises(PropertyTypeError,
            property_type1.check_value, [10, 20])
        raises(PropertyTypeError,
            property_type1.check_value, 'hello')
        raises(PropertyTypeError,
            property_type1.check_value, ['hello', 'bye'])
        raises(AssertionError,
            self.plib.get_property_type,
            PropertyType.ATOMIC, PropertyType.PHYSICAL_VALUE,
            element_sub_type=1)
        raises(AssertionError,
            self.plib.get_property_type,
            PropertyType.ATOMIC, PropertyType.PHYSICAL_VALUE, key_type=1)
        raises(AssertionError,
            self.plib.get_property_type,
            PropertyType.ATOMIC, PropertyType.PHYSICAL_VALUE, key_sub_type=1)

    def check_material_type (self):
        if get_property_db() is None:
            return
        property_type1 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.MATERIAL)

        assert property_type1.is_atomic()
        assert not property_type1.is_int()
        assert not property_type1.is_string()
        assert not property_type1.is_physical_value()
        assert property_type1.is_material()
        assert not property_type1.is_list()
        assert not property_type1.is_table()
        assert not property_type1.is_boolean()

        property_type2 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.MATERIAL)
        assert property_type2 is property_type1

        silicone = self.mdb.get_material('silicone')
        gold = self.mdb.get_material('gold')
        inch = get_standard_unit('inch')

        property_type2 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.MATERIAL)
        assert property_type2 is property_type1

        property_type1.check_value(None)
        property_type1.check_value(silicone)
        property_type1.check_value(gold)
        raises(PropertyTypeError,
            property_type1.check_value, [gold])
        raises(PropertyTypeError,
            property_type1.check_value, 'hello')
        raises(PropertyTypeError,
            property_type1.check_value, ['hello', 'bye'])
        property_type2.check_value(silicone)


    def check_list_type (self):
        if get_property_db() is None:
            return
        property_type1 = self.plib.get_property_type(
            PropertyType.LIST, PropertyType.INT)
        assert not property_type1.is_atomic()
        assert property_type1.is_int()
        assert not property_type1.is_string()
        assert not property_type1.is_physical_value()
        assert not property_type1.is_material()
        assert property_type1.is_list()
        assert not property_type1.is_table()
        assert not property_type1.is_boolean()

        property_type2 = self.plib.get_property_type(
            PropertyType.LIST, PropertyType.INT)
        assert property_type2 is property_type1
        property_type2 = self.plib.get_property_type(
            PropertyType.LIST, PropertyType.STRING)
        assert property_type2 is not property_type1
        property_type3 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.STRING)
        assert property_type3 is property_type2.element_type

        assert not property_type2.is_int()
        assert property_type2.is_string()

        property_type1.check_value(None)
        property_type1.check_value([10])
        property_type1.check_value([10, 20])
        property_type1.check_value([RangeValue(10, 20)])
        property_type1.check_value([10, 20, RangeValue(10, 20)])
        raises(PropertyTypeError,
            property_type1.check_value, 10)
        raises(PropertyTypeError,
            property_type1.check_value, RangeValue(10, 20))
        raises(PropertyTypeError,
            property_type1.check_value, 'hello')

    def check_table_type (self):
        if get_property_db() is None:
            return
        property_type1 = self.plib.get_property_type(
            PropertyType.TABLE, PropertyType.PHYSICAL_VALUE,
            key_type=PropertyType.MATERIAL)

        assert not property_type1.is_atomic()
        assert not property_type1.is_int()
        assert not property_type1.is_string()
        assert property_type1.is_physical_value()
        assert not property_type1.is_material()
        assert not property_type1.is_list()
        assert property_type1.is_table()
        assert not property_type1.is_boolean()

        assert not property_type1.is_key_int()
        assert not property_type1.is_key_string()
        assert not property_type1.is_key_physical_value()
        assert property_type1.is_key_material()

        silicone = self.mdb.get_material('silicone')
        gold = self.mdb.get_material('gold')
        inch = get_standard_unit('inch')

        property_type2 = self.plib.get_property_type(
            PropertyType.TABLE, PropertyType.PHYSICAL_VALUE,
            key_type=PropertyType.MATERIAL)
        assert property_type2 is property_type1
        property_type2 = self.plib.get_property_type(
            PropertyType.TABLE, PropertyType.PHYSICAL_VALUE,
            key_type=PropertyType.MATERIAL)
        assert property_type2 is property_type1
        property_type3 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.PHYSICAL_VALUE)
        assert property_type3 is property_type1.element_type
        assert property_type3 is property_type2.element_type
        property_type3 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.MATERIAL)
        assert property_type3 is property_type1.key_type
        assert property_type3 is property_type2.key_type
        property_type3 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.MATERIAL)
        assert property_type3 is property_type1.key_type
        assert property_type3 is property_type2.key_type
        property_type3 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.PHYSICAL_VALUE)
        assert property_type3 is property_type1.element_type
        property_type3 = self.plib.get_property_type(
            PropertyType.ATOMIC, PropertyType.MATERIAL)
        assert property_type3 is property_type1.key_type

        property_type1.check_value(None)
        property_type1.check_value(
            {gold:PhysicalValue(10),
             silicone:PhysicalValue(RangeValue(10, 20))})

        raises(PropertyTypeError,
            property_type1.check_value, 10)
        raises(PropertyTypeError,
            property_type1.check_value, [10])
        raises(PropertyTypeError,
            property_type1.check_value, RangeValue(10, 20))
        raises(PropertyTypeError,
            property_type1.check_value, 'hello')
        raises(PropertyTypeError,
            property_type1.check_value, PhysicalValue(10))
        raises(PropertyTypeError,
            property_type1.check_value, [PhysicalValue(10)])
        raises(PropertyTypeError,
            property_type1.check_value, [gold])
        raises(AssertionError,
            self.plib.get_property_type,
            PropertyType.TABLE, PropertyType.INT)

if __name__ == "__main__":
    PropertyTypeTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_unit.py $
$Id: utest_unit.py 30485 2008-01-21 21:06:41Z dbinger $
"""
from sancho.utest import UTest, raises
from dulcinea.physical_unit import PhysicalUnit, UnitCollection, UnitError
from qp.lib.spec import get_spec_problems

class PhysicalUnitTest (UTest):

    def _pre (self):
        self.units = UnitCollection()

    def _post (self):
        del self.units

    def get_units (self, *names):
        return [self.units[name] for name in names]


    def check_basics (self):
        "basic unit finding and conversion"
        units = self.units

        # Assume these units already exist.
        kilogram = units['kg']
        gram = units['g']
        metre = units['m']
        inch = units['inch']

        assert kilogram.name == "kg"
        assert kilogram.get_name() == "kg"
        assert kilogram.get_name(html=1) == "kg"

        assert kilogram.factor == 1.0
        assert kilogram.offset == 0.0
        assert not kilogram.artificial

        assert kilogram.convert_value(1.5, gram) == 1500.0
        assert round(metre.get_conversion_factor(inch), 2) == 39.37

        get_unit = units.get_unit
        assert get_unit('kg') == kilogram
        assert (kilogram*metre).artificial
        assert get_unit('kg*m').artificial
        raises(TypeError, get_unit, 66)
        assert kilogram.is_dimension('mass')
        assert metre.is_dimension('length')
        assert units['Torr'].is_dimension('pressure')
        assert get_spec_problems(units) == [], get_spec_problems(units)
        assert get_spec_problems(kilogram) == []

    def check_temperature (self):
        units = self.units
        degF = units.get_unit('degF')
        degC = units.get_unit('degC')
        kelvin = units.get_unit('K')

        warm_day = 80                   # in fahrenheit
        cold_day = 0                    # in fahrenheit
        fpw = 0                         # "freezing point of water" in celsius
        htsc = 70                       # "high temperature" superconductivity
                                        # (at a brisk 70 K)
        abs_zero = 0                    # in kelvin

        assert round(degF.convert_value(warm_day, degC),2) == 26.67
        assert round(degF.convert_value(warm_day, kelvin),2) == 299.82

        assert round(degF.convert_value(cold_day, degC),2) == -17.78
        assert round(degF.convert_value(cold_day, kelvin),2) == 255.37

        assert degC.convert_value(fpw, degC) == fpw
        assert round(degC.convert_value(fpw, degF),2) == 32.00
        assert round(degC.convert_value(fpw, kelvin),2) == 273.15

        assert round(kelvin.convert_value(htsc, degC),2) == -203.15
        assert round(kelvin.convert_value(htsc, degF),2) == -333.67

        assert round(kelvin.convert_value(abs_zero, kelvin),2) == 0
        assert round(kelvin.convert_value(abs_zero, degC),2) == -273.15
        assert round(kelvin.convert_value(abs_zero, degF),2) == -459.67

        assert degC.is_dimension('temperature')
        assert degF.is_dimension('temperature')
        assert kelvin.is_dimension('temperature')
        assert not kelvin.is_dimension('mass')


    def check_conversion (self):
        "conversion functions on unit collection"
        units = self.units
        inch = units.get_unit('inch')
        mm = units.get_unit('mm')
        degC = units.get_unit('degC')
        degF = units.get_unit('degF')
        K = units.get_unit('K')
        assert round(units.convert_value(6, inch, mm), 2) == 152.4
        assert round(units.convert_value(0, degC, degC), 2) == 0.0
        assert round(units.convert_value(0, degC, degF), 2) == 32.00
        assert round(units.convert_value(0, degC, K), 2) == 273.15

    def check_add_units (self):
        "define alternate units and convert to/from them"

        units = self.units
        units.add_unit('ft', units['inch'], 12)
        units.add_unit('yard', units['ft'], 3)

        yard = units.get_unit('yard')
        assert round(yard.get_conversion_factor(), 4) == .9144
        inch = units.get_unit('inch')
        assert inch.powers == (1, 0, 0, 0, 0, 0, 0)
        assert round(inch.factor, 5) == 0.0254

        unit = units['ft'] / units['min']
        assert unit.powers == (1, 0, -1, 0, 0, 0, 0)
        assert round(unit.factor, 5) == 0.00508
        assert round(units.convert_value(3, units['yard'],
                                         units['m']), 4) == 2.7432
        assert round(units.convert_value(500, units['s'],
                                         units['hour']), 4) == .1389

    def check_implicit (self):
        "implicit unit definition"

        units = self.units

        units.add_unit('ft', units['inch'], 12)
        units.add_unit('mile', units['ft'], 5280)
        mph = units['mile'] / units['hour']
        assert units['mile/hour'] is mph

        mps = units.get_unit('m/s', 1)
        ipm = units['inch'] / units['min']
        assert round(ipm.convert_value(1.5, mps), 7) == 6.35e-4
        assert round(ipm.convert_value(1.5, mph), 6) == 1.42e-3

    def check_compare (self):
        units = self.units
        units.get_unit("cm")
        (metre, centimetre, inch, sec, hertz) = \
            self.get_units('m', 'cm', 'inch', 's', 'Hz')

        assert metre == metre
        assert not metre == centimetre
        assert metre.is_compatible(centimetre)
        assert metre.is_compatible(inch)
        assert not metre.is_compatible(None)

        assert not sec.is_compatible(hertz)
        assert sec.is_dimension('time')
        assert sec.is_inverse_compatible(hertz)

        assert hertz.is_dimension('frequency')

        metre2 = PhysicalUnit("m", metre.powers, units)
        metre3 = PhysicalUnit("foo", metre.powers, units)
        assert metre == metre2
        assert metre == metre3

        # This depends on the list of temperature units defined in
        # UnitCollection.  There aren't that many temperature units in
        # the world, though, so it's unlikely that we'd add more and
        # break this test.
        temp_units = units.get_compatible_units('degC')
        temp_units = sorted([(u.factor, u.name) for u in temp_units])
        assert temp_units == [(5./9, 'degF'), (1, 'K'), (1, 'degC')]

        time_units = sorted([(u.factor, u.name)
                      for u in units.get_compatible_units('s')])
        expect = [(1, 's'), (60, 'min'), (60*60, 'hour'),
                  (60*60*24, 'day'), (60*60*24*7, 'week')]
        assert time_units == expect

        # Test that implicit addition affects get_compatible_units()
        units.get_unit("us")
        units.get_unit("ms")
        time_units = sorted([(u.factor, u.name)
                      for u in units.get_compatible_units('s')])
        expect = [(1e-6, "us"), (1e-3, "ms")] + expect
        assert time_units == expect


    def check_arith1 (self):
        self.units.get_unit("cm")
        (metre, centimetre, inch) = \
            self.get_units('m', 'cm', 'inch')

        scaled_cm = centimetre*100.0
        assert scaled_cm.powers == metre.powers
        assert scaled_cm.factor == metre.factor
        assert metre is scaled_cm
        assert str(scaled_cm) == "m"

        alt_cm = centimetre*1
        assert centimetre is alt_cm
        assert alt_cm*100 is metre
        assert str(alt_cm) == "cm"

        scaled_cm = centimetre * inch.get_conversion_factor(centimetre)
        assert scaled_cm.powers == inch.powers
        assert scaled_cm.factor == inch.factor
        assert inch is scaled_cm
        assert str(scaled_cm) == "inch"

    def check_arith2 (self):
        "arithmetic on physical units with other physical units"

        self.units.get_unit("um")
        (inch, um, sec, metre,) = \
               self.get_units('inch', 'um', 's', 'm')

        ips = inch/sec
        mps = metre/sec
        assert ips.get_name() == "inch/s"
        assert mps.get_name() == "m/s"
        assert ips.get_powers() == (1, 0, -1, 0, 0, 0, 0)
        assert mps.get_powers() == (1, 0, -1, 0, 0, 0, 0)
        assert ips.is_compatible(mps)
        assert round(ips.convert_value(10.0, mps),3) == 0.254

        minv = metre.inv()
        assert minv.get_name() == "1/m"
        assert minv.get_powers() == (-1, 0, 0, 0, 0, 0, 0)
        assert minv.get_conversion_factor() == 1.0

        assert (ips*metre).get_name() == "(inch/s)*m"
        assert (minv/ips).get_name() == "(1/m)/(inch/s)"
        assert (minv/4).get_name() == "1/m/4"

        umps = um/sec
        assert (umps*metre).get_name() == "(um/s)*m"
        assert (minv/umps).get_name() == "(1/m)/(um/s)"
        assert (minv/4).get_name(html=1) == "1/m/4"
        assert (minv/(1/3.0)).get_name(html=1) == "1/m/0.333333"

        msq = metre*metre
        assert msq.is_compatible(self.units['m^2'])
        assert msq.get_name() == "m*m"
        assert msq.get_powers() == (2, 0, 0, 0, 0, 0, 0)
        assert not msq.is_compatible(metre)
        raises(UnitError, msq.convert_value, 1, mps)

    def check_delete (self):
        # Can't delete standard SI units
        raises(TypeError, self.units.delete_unit, 'm')
        raises(AssertionError, self.units.delete_unit, self.units['g'])
        raises(AssertionError, self.units.delete_unit, self.units['Wb'])

        # Some deletions that should work
        cmps = self.units.get_unit('cm/s')
        assert self.units['cm/s'] == cmps
        self.units.delete_unit(cmps)
        raises(ValueError, self.units.get_unit, 'cm/s', create=False)

        # get_unit() can bring it back (but as a different object)
        assert self.units.get_unit('cm/s') == cmps
        assert self.units.get_unit('cm/s') is not cmps

        # This one can't be brought back
        mil = self.units['mil']
        self.units.delete_unit(mil)
        try:
            self.units['mil']
            assert 0
        except KeyError: pass
        raises(ValueError, self.units.get_unit, 'mil')

    def check_massquery (self):
        # Don't make any assumptions about which units are "natural" or
        # "artificial" -- it's non-obvious and a bit surprising in a
        # newly-created UnitCollection.  Doesn't really matter in real
        # life, though, since the 'artificial' flag was added well after
        # all the important units already existed in our live database.
        all_units = list(self.units.units.values())
        nat_units = self.units.get_natural_units()
        art_units = [u for u in self.units.units.values() if u.get_artificial()]
        assert set(all_units) == set(nat_units) | set(art_units)

if __name__ == "__main__":
    PhysicalUnitTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/issue.py $
$Id: issue.py 30484 2008-01-21 20:41:06Z dbinger $
"""
from datetime import timedelta
from dulcinea.base import DulcineaPersistent
from dulcinea.timestamped import timestamp_sorted, Timestamped
from dulcinea.user import DulcineaUser
from durus.persistent import PersistentObject
from durus.persistent_dict import PersistentDict
from qp.lib.spec import mapping, sequence, spec, add_getters, specify
from qp.lib.spec import require, boolean, string, Specified
from qp.pub.common import get_publisher, site_now

def get_issue_db():
    return get_publisher().get_connection().get_root().get('issue_db')


class Issue (DulcineaPersistent, Timestamped):
    """
    An Issue is usually a bug or a feature request.
    """
    title_is = (string, None)
    key_is = (string, None)
    description_is = spec(
        (string, None),
        "A description of the issue and users notes about it.")
    bang_ratings_is = spec(
        mapping({DulcineaUser:(int,None)}, PersistentDict),
        "The bang ratings given by users.")
    buck_ratings_is = spec(
        mapping({DulcineaUser:(int,None)}, PersistentDict),
        "The buck ratings given by users.")
    approvals_is = spec(
        sequence(DulcineaUser, set),
        "The users who agree the issue is resolved.")

    BANG_OPTIONS = (None, 1, 2, 3, 4, 5)
    BANG_DESCRIPTIONS = ('No Opinion',
                         'Minor',
                         'Valuable',
                         'Very Valuable',
                         'Major Advance',
                         'Essential')
    BUCK_OPTIONS = (None, 1, 2, 3, 4, 5)
    BUCK_DESCRIPTIONS = ('No Opinion',
                         'Trivial',
                         'Easy',
                         'Moderate',
                         'Tough',
                         'Extremely Difficult')

    def __init__(self):
        Timestamped.__init__(self)
        specify(self, title='', key=None, description=None,
                bang_ratings=PersistentDict(),
                buck_ratings=PersistentDict(),
                approvals=set())

    def set_key(self, key):
        assert self.key is None
        specify(self, key=key)

    def set_title(self, title):
        specify(self, title=title)

    def set_description(self, description):
        specify(self, description=description)

    def set_bang(self, user, rating):
        require(user, DulcineaUser)
        require(rating, (int, None))
        if not rating in self.BANG_OPTIONS:
            raise TypeError('invalid rating: %r, rating must be one of %r' % (
                rating, self.BANG_OPTIONS))
        self.bang_ratings[user] = rating

    def get_individual_bang(self, user):
        require(user, DulcineaUser)
        return self.bang_ratings.get(user, None)

    def get_combined_bang(self):
        cumulative_bang = 0.0
        number_of_ratings = 0
        for rating in self.bang_ratings.values():
            if rating is not None:
                cumulative_bang += rating
                number_of_ratings += 1
        if number_of_ratings == 0:
            return None
        else:
            return cumulative_bang / number_of_ratings

    def get_bang_description(self, rating):
        for index, option in enumerate(self.BANG_OPTIONS):
            if rating <= option:
                return self.BANG_DESCRIPTIONS[index]

    def set_buck(self, user, rating):
        require(user, DulcineaUser)
        require(rating, (int, None))
        if rating not in self.BUCK_OPTIONS:
            raise TypeError('invalid rating: %r, rating must be one of %r' % (
                rating, self.BUCK_OPTIONS))
        self.buck_ratings[user] = rating

    def get_individual_buck(self, user):
        require(user, DulcineaUser)
        return self.buck_ratings.get(user, None)

    def get_combined_buck(self):
        cumulative_buck = 0.0
        number_of_ratings = 0
        for rating in self.buck_ratings.values():
            if rating is not None:
                cumulative_buck += rating
                number_of_ratings += 1
        if number_of_ratings == 0:
            return None
        else:
            return cumulative_buck / number_of_ratings

    def get_buck_description(self, rating):
        for index, option in enumerate(self.BUCK_OPTIONS):
            if rating <= option:
                return self.BUCK_DESCRIPTIONS[index]

    def get_priority(self):
        bang = self.get_combined_bang()
        if bang is None:
            return None
        buck = self.get_combined_buck()
        if buck is None:
            return None
        return bang / buck

    def get_inverted_priority(self):
        priority = self.get_priority()
        if priority is None:
            return 0
        else:
            return 1 / priority

    def set_approval(self, user, approval):
        require(user, DulcineaUser)
        require(approval, boolean)
        self._p_note_change()
        if approval:
            self.approvals.add(user)
        elif user in self.approvals:
            self.approvals.remove(user)

    def awaits_approval_by(self, user):
        require(user, DulcineaUser)
        return (self.approvals and
                user in self.bang_ratings.keys() and
                user not in self.approvals)

    def is_approved_by(self, user):
        require(user, DulcineaUser)
        return user in self.approvals

    def get_voters(self):
        return list(self.bang_ratings.keys())

    def is_resolved(self):
        for user in self.bang_ratings.keys():
            if user not in self.approvals:
                return False
        return True

    def get_approvers(self):
        return self.approvals

add_getters(Issue)

def _sort_issue_list(issues):
    def issue_info(issue):
        return (len(issue.get_approvers()), issue.get_inverted_priority())
    return sorted(issues, key=issue_info)


class IssueDatabase (PersistentObject, Specified):

    issues_is = spec(
        mapping({string:Issue}, PersistentDict),
        "Mapping of issue IDs to issues.")
    issues_in_progress_is = spec(
        mapping({DulcineaUser:Issue}, PersistentDict),
        "Mapping of users to the issues they are working on.")
    _next_issue_id_is = spec(
        int,
        "Internal counter used for allocating issue IDs")

    def __init__(self):
        specify(self,
                issues=PersistentDict(),
                issues_in_progress=PersistentDict(),
                _next_issue_id=1)

    def get_issue(self, issue_id):
        return self.issues.get(issue_id)

    def __getitem__(self, issue_id):
        return self.issues[issue_id]

    def _generate_issue_id(self):
        issue_id = self._next_issue_id
        self._next_issue_id += 1
        return 'I%04d' % issue_id

    def add_issue(self, issue):
        require(issue, Issue)
        assert issue.get_key() is None, 'Issue already has id'
        issue_id = self._generate_issue_id()
        assert not self.issues.has_key(issue_id)
        issue.set_key(issue_id)
        self.issues[issue_id] = issue

    def remove_issue(self, issue):
        issue_id = issue.get_key()
        assert issue_id is not None, 'Issue has no ID'
        del self.issues[issue_id]
        for user, issue_in_progress in self.issues_in_progress.items():
            if issue is issue_in_progress:
                del self.issues_in_progress[user]

    def get_issues(self):
        return _sort_issue_list(self.issues.values())

    def set_issue_in_progress(self, user, issue):
        require(user, DulcineaUser)
        require(issue, Issue)
        self.issues_in_progress[user] = issue

    def clear_issue_in_progress(self, user, issue):
        require(user, DulcineaUser)
        require(issue, Issue)
        if self.issues_in_progress.get(user) is issue:
            del self.issues_in_progress[user]

    def get_issue_in_progress(self, user):
        require(user, DulcineaUser)
        return self.issues_in_progress.get(user)

    def get_issues_in_progress(self):
        return _sort_issue_list(set(self.issues_in_progress.values()))

    def is_issue_in_progress(self, issue):
        require(issue, Issue)
        return issue in self.issues_in_progress.values()

    def get_users_on_issue(self, issue):
        require(issue, Issue)
        return [user
                for user, issue_in_progress in self.issues_in_progress.items()
                if issue is issue_in_progress]

    def get_issues_for_approval(self, user):
        require(user, DulcineaUser)
        return _sort_issue_list([issue
                                 for issue in self.issues.values()
                                 if issue.awaits_approval_by(user)])

    def get_recent_issues(self, days):
        minimum_time = site_now() - timedelta(days=days)
        issues = [issue for issue in self.issues.values()
                  if issue.get_timestamp() > minimum_time]
        return timestamp_sorted(issues)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/stored_file.py $
$Id: stored_file.py 31151 2008-09-18 11:18:33Z dbinger $

Store a file in the filesystem.
"""
import os, grp, sys
import errno
from dulcinea.user import DulcineaUser
from durus.persistent import PersistentObject
from qp.lib.util import rand_str
from qp.lib.spec import spec, specify, add_getters_and_setters, either
from qp.lib.spec import string, Specified, datetime_with_tz
from qp.pub.common import get_config_value, site_now

digits = '0123456789'
lowercase = 'abcdefghijklmnopqrstuvwxyz'

ALLOWABLE_SPECIAL_CHARS = "+-_.@"

def path_check(path):
    """(path:string) -> string | None

    Return a lower-case and normalized version of the 'path' after
    checking for allowable characters.
    Return None if path has '..' or any special characters
    other than what's in lowercase, digits, and
    the ALLOWABLE_SPECIAL_CHARS string
    """
    if '..' in path:
        return None
    path = path.lower()
    for letter in path:
        if (letter != '/' and letter not in lowercase and
            letter not in digits and
            letter not in ALLOWABLE_SPECIAL_CHARS):
            return None
    return path

MIME_TYPES = [("image/gif", "GIF image"),
              ("image/jpeg", "JPEG image"),
              ("image/png", "PNG image"),
              ("image/tiff", "TIFF image"),
              ("image/x-ms-bmp", "BMP image"),
              ("text/plain", "plain text"),
              ("text/html", "HTML text"),
              ("text/rtf", "richtext document"),
              ("application/pdf", "PDF document"),
              ("application/msword", "Word document"),
              ("application/vnd.ms-powerpoint", "PowerPoint document"),
              ("application/vnd.ms-excel", "Excel document"),
              ("application/postscript", "PostScript file"),
              ("application/octet-stream", "Binary data"),
              ("application/x-gtar", "gzipped tar file"),
              ("application/x-gzip", "gzipped file"),
              ("application/x-tar", "tar file"),
              ("application/zip", "zip file"),
              ("video/x-msvideo", ".avi file"),              
              ("video/quicktime", ".mov file"),                            
              ]

def get_file_store():
    return get_config_value('file_store', '/var/tmp')


class StoredFile (PersistentObject, Specified):
    """
    Class for reading from or writing to a stored file

    A StoredFile is a wrapper around a file in a place in the
    filesystem known as the 'file_store'.

    StoredFiles can be instantiated with a fresh source file to move a new
    file into the file_store, or instantiated by referencing an existing
    file in the file_store.
    """
    path_is = spec(
        string,
        "Path to the physical file containing the stored file's data.")
    filename_is = spec(
        string,
        "Filename to use when downloading this file")
    mime_type_is = spec(
        either(None, *[pair[0] for pair in MIME_TYPES]),
        "MIME type of the contents of this file.")
    charset_is = spec(
        (None, string),
        "The charset encoding of the file, if known.")
    description_is = spec(
        (string, None),
        "Description of this file.")
    owner_is = spec(
        (DulcineaUser, None),
        "The file's owner")
    date_is = spec(
        datetime_with_tz,
        "timestamp")

    def __init__(self, path):
        path = os.path.normpath(path)
        self.path = path_check(path)
        assert self.path, 'Bad name: %r' % path
        specify(self,
                description=None,
                owner=None,
                date=site_now(),
                filename=os.path.basename(self.path),
                mime_type=None,
                charset=None)

    def get_id(self):
        return self.path.replace('/', '')

    def open(self):
        """() -> file
        Returns a Python file object opened for reading.  This can be used to
        read the file's content.
        """
        return open(self.get_full_path(), 'rb')

    def chgrp(self, grp_id):
        """(grp_id : string | int)
        Change the category ownership of this file to the specified ID.
        """
        if isinstance(grp_id, string):
            grnam, grpass, gid, grmembers = grp.getgrnam(grp_id)
            grp_id = gid
        os.chown(self.get_full_path(), -1, grp_id)

    def set_mime_type(self, mime_type):
        """(mime_type:string)
        Sets the MIME type for the file.
        """
        # Check there's a single '/' in the MIME type
        if mime_type.count('/') != 1:
            raise ValueError("Invalid MIME type %r" % mime_type)
        specify(self, mime_type = mime_type)

    def get_size(self):
        """() -> int
        Return the size of the file, measured in bytes, or None if
        the file doesn't exist.
        """
        path = self.get_full_path()
        if not os.path.exists(path):
            return None
        stats = os.stat(path)
        return stats.st_size

    def get_full_path(self):
        """() -> string
        """
        return os.path.join(get_file_store(), self.path)

    def has_manage_access(self, user):
        return (user is self.owner or user.is_admin())

add_getters_and_setters(StoredFile)


def get_random_path():
    h = rand_str(6)
    return '%s/%s/%s' % (h[-1], h[-3:-1], h[0:-3])

def new_file(fp):
    root_directory = get_file_store()
    flags = os.O_WRONLY|os.O_CREAT|os.O_EXCL
    try:
        flags |= os.O_BINARY    # for Windows
    except AttributeError:
        pass
    for attempt in range(1000):
        path = get_random_path()
        full_path = os.path.join(root_directory, path)
        destdir = os.path.dirname(full_path)
        if not os.path.exists(destdir):
            os.makedirs(destdir, mode=int("775", 8))
        try:
            fd = os.open(full_path, flags)
        except OSError:
            exc = sys.exc_info()[1]
            if exc.errno == errno.EEXIST:
                pass # destination exists, retry
            else:
                raise # some other error
        else:
            break
    else:
        raise RuntimeError("Unable to make temp file.")
    dest_fp = os.fdopen(fd, "wb")
    while 1:
        chunk = fp.read(10000)
        if not chunk:
            break
        dest_fp.write(chunk)
    fp.close()
    dest_fp.close()
    return StoredFile(path)




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_history.py $
$Id: utest_history.py 31124 2008-09-16 17:35:28Z dbinger $
"""
from datetime import timedelta
from dulcinea.base import DulcineaPersistent
from dulcinea.common import CommonTest
from dulcinea.history import History, Event, Historical
from dulcinea.user import DulcineaUser
from qp.pub.common import site_now
import time


class HistoryTest (CommonTest):

    def check_historical(self):
        class PHistorical(DulcineaPersistent, Historical):
            pass
        PHistorical().get_history()

    def check_event(self):
        user = DulcineaUser("jblow")
        message = "created new foobar ding-dong"
        event1 = Event(user, "create", message)
        assert event1.get_user() == user
        assert event1.get_event_code() == 'create'
        assert event1.get_message() == message

        assert str(event1) == "%s create" % event1.timestamp

        event2 = Event(user, "modify", "changed blah blah on foobar")
        assert event2.timestamp >= event1.timestamp
        time.sleep(0.01)
        event3 = Event(user, "delete", "jblow deleted foobar")
        assert event3.timestamp > event2.timestamp
        assert event3.timestamp is event3.get_timestamp()

    def check_history(self):
        history = History()
        user1 = DulcineaUser("joe")
        user2 = DulcineaUser("jane")        
        assert len(history) == 0
        assert str(history) == "history (0 events)"
        history.add_event(user1, 'create', 'joe created blah')
        history.add_event(None, 'modify', 'jane modified blah')
        assert len(history) == 2
        assert history[0].timestamp <= history[1].timestamp
        assert history[0].user == user1
        assert history[1].user == None

        history.add_event(user2, 'create', 'a')
        assert len(history) == 3
        assert (history[0].timestamp <
                history[1].timestamp <=
                history[2].timestamp)
        history.add_event(None, 'create', 'modern age begins')
        assert len(history) == 4
        assert history[0].message == "joe created blah"
        assert history[1].message == "jane modified blah"


    def check_variations(self):
        history = History()
        # Works with no user object
        history.add_event(None, 'modify', 'tweaked it')
        assert history[-1].user == None

        # Replacing existing event
        history.update_last_event(None, 'modify', 'tweaked it more')
        assert history[-1].message == 'tweaked it more'

        history.update_last_event(None, 'create', 'creation')
        assert history[-2].message == 'tweaked it more'

    def check_find(self):
        user1 = DulcineaUser("joe")
        user2 = DulcineaUser("jane")
        history = History()
        assert history.find_event() == None
        assert history.find_event(latest=True) == None

        history.add_event(user1, "create", "joe created it")
        assert history.find_event().message == "joe created it"
        assert history.find_event(latest=True).message == "joe created it"
        assert history.find_event(user=user1).message == "joe created it"
        assert history.find_event(user=user2) == None
        assert history.find_event(event_code='create').message == (
            "joe created it")
        assert history.find_event(message='no created it') == None

        assert history.find_event(event_code='access') == None
        assert history.find_event(event_codes=['access']) == None

        start = site_now() - timedelta(1)
        history.add_event(user2, "access", "jane accessed it")
        history.add_event(user1, "modify", "joe modified it")
        history.add_event(user2, "modify", "jane modified it")
        history.add_event(user2, "delete", "jane deleted it")

        assert history.find_event().message == "joe created it"
        assert history.find_event(latest=True).message == "jane deleted it"
        assert history.find_event(user=user1,latest=True).message == (
            "joe modified it")
        assert history.find_event(event_code='modify').message == (
            "joe modified it")
        assert history.find_event(event_code='modify',latest=True).message == (
            "jane modified it")
        assert len(history.find_events_after(start)) == 5
        assert history.find_last_date('foo') == None


if __name__ == "__main__":
    HistoryTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_country.py $
$Id: utest_country.py 27310 2005-09-06 18:03:40Z dbinger $
"""
from sancho.utest import UTest, raises
from dulcinea.country import get_country_name, get_country_list, \
     get_country_codes

class CountryTest (UTest):

    def check_country_lookup(self):
        assert get_country_name('CA') == "Canada"
        assert get_country_name('AU') == "Australia"
        assert get_country_name('TV') == "Tuvalu"
        raises(KeyError, get_country_name, 'XX')
        raises(KeyError, get_country_name, None)

    def check_country_list(self):
        list = get_country_list()
        assert len(list) > 200
        assert ('FI', 'Finland') in list
        assert ('BR', 'Brazil') in list
        assert 0 not in [type(x) is tuple and len(x) == 2 for x in list]

    def check_country_codes(self):
        assert get_country_codes()

if __name__ == "__main__":
    CountryTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/session.py $
$Id: session.py 29688 2007-03-29 13:46:15Z dbinger $
"""
from dulcinea.attachable import Attachable
from dulcinea.user import DulcineaUser
from qp.lib.spec import Specified
from qp.pub.session import Session


class DulcineaSession (Session, Specified, Attachable):

    effective_user_is = DulcineaUser
    owner_is = DulcineaUser

    def __init__(self):
        Session.__init__(self)
        Attachable.__init__(self)

    def clear_app_state(self):
        """Override to clear any session data
        """



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_pvalue.py $
$Id: utest_pvalue.py 31120 2008-09-16 16:23:21Z dbinger $
"""
from dulcinea.physical_unit import UnitCollection
from dulcinea.physical_value import PhysicalValue, UnitError
from dulcinea.range_value import RangeValue
from sancho.utest import UTest


class PhysicalValueTest (UTest):

    def _pre (self):
        self.units = units = UnitCollection()
        units.add_unit('ft', units['inch'], 12)
        units.add_unit('km', units['m'], 1000)
        units.add_unit('mile', units['ft'], 5280)
        units.add_unit('L', units['m^3'], 0.1**3)
        units.add_unit('galUS', units['L'], 3.785411784)
        meters = units.get_unit('m')
        self.len1 = PhysicalValue(1.0, meters)
        self.len1a = PhysicalValue(1, meters)
        self.len2 = PhysicalValue(2.0, meters)
        self.len3 = PhysicalValue(3, units['ft'])
        self.len4 = PhysicalValue(36, units['inch'])
        self.len5 = PhysicalValue(RangeValue(3.0, 5.0), units['cm'])

        self.time1 = PhysicalValue(15, units['s'])
        self.time2 = PhysicalValue(0.5, units['min'])
        self.nounit1 = PhysicalValue(0.5)
        self.nounit2 = PhysicalValue(1.0)

    def check_basics(self):
        meters = self.units.get_unit('m')
        (len1, len1a, len2, len5) = \
            (self.len1, self.len1a, self.len2, self.len5)
        assert len1.unit is len2.unit
        assert len1.get_unit() is len2.get_unit()
        assert len2.get_unit() == meters
        assert len2.get_unit_name() == "m"
        assert len1.value == 1.0
        assert len1.get_value() == 1.0
        assert len2.value == 2.0
        assert len2.get_value() == 2.0
        assert self.nounit1.get_unit_name() == None
        assert len1.get_type() == float
        assert len1a.get_type() == float
        assert len5.get_type() == RangeValue
        try:
            PhysicalValue(RangeValue(1, 2.0))
            assert 0
        except TypeError: pass
        try:
            PhysicalValue( (1,2) )
            assert 0
        except TypeError: pass

    def check_format(self):
        (nounit1, len4, len5) = (self.nounit1, self.len4, self.len5)
        assert str(nounit1) == "0.5"
        assert nounit1.format() == "0.5"
        assert str(len4) == "36.0 inch"
        assert len4.format() == "36 inch"
        assert str(len5) == "3.0 .. 5.0 cm"
        assert len5.format() == "3 .. 5 cm"
        assert nounit1.format(html=1) == "0.5"
        assert nounit1.format(show_unit=0) == "0.5"
        assert len4.format(html=1) == "36&nbsp;inch"
        assert len4.format(show_unit=0) == "36"
        assert len4.format(show_unit=0, html=1) == "36"
        self.units.get_unit("um").set_name("um", "&micro;m")
        a = PhysicalValue(2, self.units.get_unit("um"))
        assert a.format() == "2 um"
        assert a.format(html=1) == "2&nbsp;&micro;m"
        b = a/3
        assert b.format() == "0.666667 um"
        assert b.format(html=1) == "0.666667&nbsp;&micro;m"
        c = a*4
        assert c.format() == "8 um"
        assert c.format(html=1) == "8&nbsp;&micro;m"
        d = a/1
        assert d.format() == "2 um"
        assert d.format(html=1) == "2&nbsp;&micro;m"
        e = a*1
        assert e.format() == "2 um"
        assert e.format(html=1) == "2&nbsp;&micro;m"
        pv = PhysicalValue(32078.1, self.units.get_unit("Ang"))
        assert pv.format() == "32078.1 Ang"
        assert pv.format(html=1) == "32078.1&nbsp;&Aring;"
        pv = pv.convert(self.units.get_unit("m"))
        assert pv.format() == "3.20781e-06 m"
        assert pv.format(html=1) == "3.20781e-06&nbsp;m"

    def _check_compatible(self):
        meters = self.units.get_unit('m')
        sec = self.units.get_unit('s')
        mps = meters/sec
        (len1, len5) = (self.len1, self.len5)
        t1 = PhysicalValue(30, sec)
        assert len1.unit_compatible(meters)
        assert len5.unit_compatible(meters)
        assert len1.unit_compatible(len5)
        assert t1.unit_compatible(self.units.get_unit('hour'))
        assert t1.unit_compatible(t1)
        assert (len1/t1).unit_compatible(mps)
        assert not len1.unit_compatible(meters/sec)
        assert not t1.unit_compatible(meters)
        nodim = meters/meters
        assert not self.nounit1.unit_compatible(meters)
        assert not self.nounit1.unit_compatible(len1)
        assert self.nounit1.unit_compatible(self.nounit2)
        assert self.nounit1.unit_compatible(None)
        assert self.nounit1.unit_compatible(nodim)

    def check_dimension(self):
        assert self.len1.is_dimension('length')
        assert not self.len1.is_dimension('time')
        assert not self.nounit1.is_dimension('length')

    def _check_cmp(self):
        (len1, len1a, len2) = (self.len1, self.len1a, self.len2)
        # compare "1.0 m" and "2.0 m"
        assert len1 != len2
        assert len1 < len2
        assert len1 <= len2
        assert len2 > len1
        assert len2 >= len1
        # compare "1.0 m" and "1 m" (float vs int)
        assert len1 == len1a
        assert len1 <= len1a
        assert len1 >= len1a
        # compare compatible but different units
        (time1, time2) = (self.time1, self.time2)
        assert time1 < time2
        assert time2 > time1
        assert time1 != time2
        time2.value = 0.25            # 0.25 min = 15 s
        assert time1 == time2
        assert time2 == time1
        assert not time1 < time2
        assert not time1 > time2
        assert time1 <= time2
        assert time1 >= time2
        # sign differences (-1 m != 1 m)
        len1_neg = PhysicalValue(-self.len1.value, self.len1.unit)
        assert len1 != len1_neg
        assert len1_neg != len1
        assert len1_neg == len1_neg
        # compare unitless PhysicalValues
        (nounit1, nounit2) = (self.nounit1, self.nounit2)
        assert nounit1 < nounit2
        assert nounit2 > nounit1
        assert not nounit2 == nounit1
        assert nounit1 == 0.5
        assert nounit1 != 25
        assert nounit1 != 'bogus'
        assert nounit1 > 0.4
        assert nounit2 < 2
        # nonsensical comparisons (units vs no units, or incompatible units)
        assert len1 != 1
        assert len2 != 0.0
        assert not len1 == nounit1
        assert not len2 == time1
        try:
            len1 < 1
            assert 0
        except UnitError: pass
        try:
            len2 >= 0.0
            assert 0
        except UnitError: pass
        try:
            len1 < nounit1
            assert 0
        except UnitError: pass
        try:
            len2 > 'bogus'
            assert 0
        except UnitError: pass

    def _check_cmp_roundoff(self):
        # compare values that expose round-off errors
        pv1 = PhysicalValue(100, self.units.get_unit("um"))
        pv2 = PhysicalValue(100000, self.units.get_unit("nm"))
        # these all require conversion; which direction the conversion
        # goes determines which are susceptible to round-off problems!
        assert pv1 == pv2
        assert pv2 == pv1
        assert pv1 <= pv2
        assert pv2 <= pv1
        assert pv1 >= pv2
        assert pv2 >= pv1
        pv3 = pv1.convert(pv2.unit)
        pv4 = pv2.convert(pv1.unit)
        # these should require no conversion: the arithmetic has been done
        assert pv3 == pv2
        assert pv2 == pv3
        assert pv4 == pv1
        assert pv1 == pv4
        # these require even more conversions (eg. *back* to original
        # units), so the round-off should be even worse
        assert pv3 == pv1
        assert pv1 == pv3
        assert pv4 == pv2
        assert pv2 == pv4
        assert pv3 == pv4
        assert pv4 == pv3

    def _check_convert(self):
        (len1, len1a, len2) = (self.len1, self.len1a, self.len2)
        units = self.units
        m = units.get_unit('m')
        s = units.get_unit('s')
        cm = units.get_unit('cm')
        inch = units.get_unit('inch')
        # convert to same unit (m -> m)
        assert len1a.convert(m).get_tuple() == (1, m)
        # convert m -> cm
        assert len1.convert(cm).get_tuple() == (100, cm)
        # convert m -> inch
        metre = len1a.convert(inch)
        assert round(metre.get_value(), 2) == 39.37
        # error case: m -> sec
        try:
            len1.convert(s)
            assert 0
        except UnitError: pass

    def check_add(self):
        foot = self.units.get_unit('ft')
        inch = self.units.get_unit('inch')
        (len1, len1a, len2, len3, len4) = \
            (self.len1, self.len1a, self.len2, self.len3, self.len4)
        # simple addition/subtraction (identical units)
        assert len1 + len1 == len2
        assert len1 + len1a == len2
        assert len2 - len1 == len1
        assert len2 - len1a == len1
        # compatible units -- make sure it preserves the unit of the
        # first one!
        assert (len3 + len4).get_tuple() == (6, foot)
        assert (len4 + len3).get_tuple() == (72, inch)
        assert round((len3 - len4).get_value(), 4) == 0
        assert round((len4 - len3).get_value(), 4) == 0
        # error case: incompatible units
        time1 = self.time1
        try:
            len1 + time1
            assert 0
        except UnitError: pass
        try:
            time1 - len1
            assert 0
        except UnitError: pass
        # two unitless PhysicalValues
        (nounit1, nounit2) = (self.nounit1, self.nounit2)
        assert nounit1 + nounit2 == PhysicalValue(1.5)
        assert nounit1 - nounit2 == PhysicalValue(-0.5)
        # unitless PhysicalValue with a regular number (and vice-versa)
        assert nounit1 + 1 == PhysicalValue(1.5)
        assert nounit1 - 1 == PhysicalValue(-0.5)
        assert 1 + nounit1 == PhysicalValue(1.5)
        assert 1 - nounit1 == PhysicalValue(0.5)
        # unitless PhysicalValue with a RangeValue
        rv = RangeValue(3.0, 4.5)
        assert nounit1 + rv == PhysicalValue(RangeValue(3.5, 5.0))
        assert rv + nounit1 == PhysicalValue(RangeValue(3.5, 5.0))
        assert nounit1 - rv == PhysicalValue(RangeValue(-4.0, -2.5))
        assert rv - nounit1 == PhysicalValue(RangeValue(2.5, 4.0))
        # error case: PhysicalValue-with-unit with a unitless
        # PhysicalValue
        try:
            len1 + nounit1
            assert 0
        except UnitError: pass
        try:
            nounit1 + len1
            assert 0
        except UnitError: pass
        # error case: PhysicalValue-with-unit with a regular number
        try:
            len1 + 2.0
            assert 0
        except UnitError: pass
        try:
            2.0 + len1
            assert 0
        except UnitError: pass
        try:
            2.0 - len1
            assert 0
        except UnitError: pass
        # error case: PhysicalValue with some random object (not another
        # PV, not a simple number, not a range)
        try:
            len1 + 'foo'
            assert 0
        except TypeError: pass

    def _check_mult(self):
        units = self.units
        m = units.get_unit('m')
        s = units.get_unit('s')
        min = units.get_unit('min')
        msq = units.get_unit('m*m', 1)
        foot = units.get_unit('ft')
        fsq = units['ft'] * units['ft']
        ftm = foot * m                  # "foot-meters"? whatever...
        ftinch = units.get_unit('ft') * units.get_unit('inch')
        minv = units.create_unit("minv", None, (-1, 0, 0, 0, 0, 0, 0), 1.0)
        (len1, len2, len3, len4) = (self.len1, self.len2, self.len3, self.len4)
        (time1, time2) = (self.time1, self.time2)
        # identical units
        assert (len1 * len2).get_tuple() == (2, msq)
        assert (len2 * len2).get_tuple() == (4, msq)
        assert (len1 * len2).get_tuple() == (2, msq)
        assert (len1 / len1).get_tuple() == (1, None)
        assert (len1 / len2).get_tuple() == (0.5, None)
        # compatible units: 1 m * 3 ft = 3 ft*m = .9144 m^2 = 9.84 ft^2
        area = len1 * len3
        assert area.get_tuple() == (3, ftm)
        assert round(area.convert(msq).get_value(),4) == .9144
        assert round(area.convert(fsq).get_value(),4) == 9.8425
        # 1 m / 3 ft = 1 m / .9144 m = 1.0936 (dimensionless quantity)
        assert round((len1 / len3).get_value(), 4) == 1.0936
        # incompatible units: the dimensions just pile up
        assert (len2 * time1).get_tuple() == (2.0*15, m*s)
        assert (len2 / time1).get_tuple() == (2.0/15, m/s)
        assert round((len2 / time1).convert(foot/min).get_value(),
                     4) == 26.2467
        # Combine unitless and "unitfull" PhysicalValue
        (nounit1, nounit2) = (self.nounit1, self.nounit2)
        assert (nounit1 * len2).get_tuple() == (1.0, m)
        assert (len2 * nounit1).get_tuple() == (1.0, m)
        assert (nounit1 / len2).get_tuple() == (0.25, minv)
        assert (len2 / nounit1).get_tuple() == (4.0, m)
        # Ditto, but now use a regular number in place of a unitless
        # PhysicalValue
        assert (0.5 * len2).get_tuple() == (1.0, m)
        assert (len2 * 0.5).get_tuple() == (1.0, m)
        assert (0.5 / len2).get_tuple() == (0.25, minv)
        assert (len2 / 0.5).get_tuple() == (4.0, m)
        rv = RangeValue(1.5, 2.5)
        assert (len1 * rv).get_tuple() == (RangeValue(1.5, 2.5), m)
        assert (rv * len1).get_tuple() == (RangeValue(1.5, 2.5), m)
        try:
            len1 / rv
            assert 0
        except TypeError: pass
        assert (rv / len1).get_tuple() == (RangeValue(1.5, 2.5), minv)
        try:
            len1 * 'foo'
            assert 0
        except TypeError: pass


if __name__ == "__main__":
    PhysicalValueTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_stored_file.py $
$Id: utest_stored_file.py 31156 2008-09-18 15:54:17Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.stored_file import path_check, new_file
from dulcinea.user import DulcineaUser
from durus.utils import as_bytes
from qp.pub.common import site_now
from sancho.utest import raises
import os, shutil

TEST_ROOT_DIR = '/tmp/storedfiletest'
TEST_CONTENT = as_bytes("Contents of a file go here.")

import dulcinea.stored_file
def get_file_store():
    return TEST_ROOT_DIR
dulcinea.stored_file.get_file_store = get_file_store

assert dulcinea.stored_file.get_file_store() == TEST_ROOT_DIR

class StoredFileTest (CommonTest):

    def check_file(self):
        if os.path.exists(TEST_ROOT_DIR):
            shutil.rmtree(TEST_ROOT_DIR)
        os.mkdir(TEST_ROOT_DIR)
        user = DulcineaUser('a')
        content_filename = os.path.join(TEST_ROOT_DIR, 'test_content')
        content = open(content_filename, 'wb')
        content.write(TEST_CONTENT)
        content.close()
        sf = new_file(open(content_filename, 'rb'))
        f = sf.open()
        assert f.read() == TEST_CONTENT
        f.close()
        assert sf.get_size() == len(TEST_CONTENT)
        raises(ValueError, sf.set_mime_type, "textplain")
        sf.set_mime_type("text/plain")
        sf.set_filename("foo")
        assert sf.get_filename() == 'foo'
        sf.set_description("boo")
        assert sf.get_description() == 'boo'
        u = DulcineaUser(None)
        sf.set_owner(u)
        assert sf.get_owner() == u
        assert sf.has_manage_access(u)
        adate = site_now()
        sf.set_date(adate)
        assert sf.get_date() == adate
        shutil.rmtree(TEST_ROOT_DIR)
        del content_filename

    def check_Path_check(self):
        assert path_check("Abc") == 'abc'
        assert path_check("abc/../../../etc/passwd") == None
        assert path_check("/special/characters/-_@.cif") == (
                      "/special/characters/-_@.cif")
        assert path_check("/illegal/characters\001/") == None


if __name__ == "__main__":
    StoredFileTest()



#!/usr/bin/env python
#
# $URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/bin/bad_links.py $
# $Id: bad_links.py 31107 2008-09-16 13:57:47Z dbinger $

import sys
from qp.lib.site import Site
from dulcinea.util import urlopen

def try_url(link):
    try:
        file = urlopen(link.get_link_url())
    except IOError:
        err = sys.exc_info()[1]
        sys.stderr.write("Link %d Unable to fetch %s: %s\n" % (
            link.get_key(), link.get_link_url(), err.strerror))

def main (prog, args):
    usage = "usage: %s [site]" % prog

    if len(args) != 1:
        sys.exit(usage)

    link_db = Site(args[0]).get_connection().get_root()['link_db']
    for link in link_db.get_all_links():
        try_url(link)

if __name__ == '__main__':
    main(sys.argv[0], sys.argv[1:])
